import ipaddress
from datetime import datetime
from typing import Optional, List, Any, Dict, Union

from pydantic import Field, validator

from glQiwiApi.types.base import Base, ExtraBase
from glQiwiApi.types.amount import CurrencyAmount, CurrencyModel
from glQiwiApi.utils.currency_util import Currency


class PassInfo(Base):
    """object: PassInfo"""

    last_pass_change: str = Field(alias="lastPassChange")
    next_pass_change: str = Field(alias="nextPassChange")
    password_used: bool = Field(alias="passwordUsed")


class MobilePinInfo(Base):
    """object: MobilePinInfo"""

    last_mobile_pin_change: Optional[datetime] = Field(alias="lastMobilePinChange")
    mobile_pin_used: bool = Field(alias="mobilePinUsed")
    next_mobile_pin_change: str = Field(alias="nextMobilePinChange")


class PinInfo(Base):
    """object: PinInfo"""

    pin_used: bool = Field(alias="pinUsed")


class AuthInfo(Base):
    """object: AuthInfo"""

    ip: ipaddress.IPv4Address
    bound_email: Optional[str] = Field(None, alias="boundEmail")
    last_login_date: Optional[datetime] = Field(None, alias="lastLoginDate")
    email_settings: Optional[Dict[Any, Any]] = Field(None, alias="emailSettings")
    mobile_pin_info: MobilePinInfo = Field(alias="mobilePinInfo")
    pass_info: PassInfo = Field(alias="passInfo")
    person_id: int = Field(alias="personId")
    pin_info: PinInfo = Field(alias="pinInfo")
    registration_date: datetime = Field(alias="registrationDate")


class SmsNotification(Base):
    """object: SmsNotification"""

    price: CurrencyAmount
    enabled: bool
    active: bool
    end_date: Optional[datetime] = Field(None, alias="endDate")


class IdentificationInfo(Base):
    """object: IdentificationInfo"""

    bank_alias: Optional[str] = Field(None, alias="bankAlias")
    identification_level: Optional[str] = Field(None, alias="identificationLevel")
    passport_expired: Optional[bool] = Field(None, alias="passportExpired")


class NickName(Base):
    """object: NickName"""

    nickname: Optional[str] = None
    can_change: bool = Field(alias="canChange")
    can_use: bool = Field(alias="canUse")
    description: str = ""


class Feature(Base):
    """object: Feature"""

    feature_id: int = Field(alias="featureId")
    feature_value: str = Field(alias="featureValue")
    start_date: str = Field(alias="startDate")
    end_date: str = Field(alias="endDate")


class ContractInfo(Base):
    """object: ContractInfo"""

    blocked: bool = False
    contract_id: int = Field(alias="contractId")
    creation_date: datetime = Field(alias="creationDate")
    identification_info: List[IdentificationInfo] = Field(alias="identificationInfo")
    sms_notification: SmsNotification = Field(alias="smsNotification")
    nickname: NickName
    features: Optional[List[Feature]] = None


class UserInfo(Base):
    """object: UserInfo"""

    default_pay_currency: CurrencyModel = Field(..., alias="defaultPayCurrency")
    default_pay_source: Optional[int] = Field(None, alias="defaultPaySource")
    default_pay_account_alias: Optional[str] = Field(
        None, alias="defaultPayAccountAlias"
    )
    email: Optional[str] = None
    first_transaction_id: Union[str, int, None] = Field(None, alias="firstTxnId")
    language: str
    operator: str
    phone_hash: str = Field(alias="phoneHash")
    promo_enabled: Optional[bool] = Field(None, alias="promoEnabled")

    @validator("default_pay_currency", pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))


class QiwiAccountInfo(ExtraBase):
    auth_info: Optional[AuthInfo] = Field(None, alias="authInfo")
    contract_info: Optional[ContractInfo] = Field(None, alias="contractInfo")
    user_info: Optional[UserInfo] = Field(None, alias="userInfo")


__all__ = ["QiwiAccountInfo"]



"""
Gracefully and lightweight wrapper to deal with QIWI API
It's an open-source project so you always can improve the quality of code/API by
adding something of your own...
Easy to integrate to Telegram bot, which was written on aiogram or another async/sync library.

"""
from __future__ import annotations

import uuid
from contextlib import suppress
from copy import deepcopy
from datetime import datetime
from typing import (
    List,
    Tuple,
    Dict,
    Union,
    Optional,
    Any,
    cast,
    Iterable,
    Type,
)

from pydantic import ValidationError

from glQiwiApi.core import RequestService, constants
from glQiwiApi.core.abc.wrapper import Wrapper
from glQiwiApi.core.constants import NO_CACHING
from glQiwiApi.core.mixins import (
    ContextInstanceMixin,
    DataMixin,
    HandlerCollectionMixin,
)
from glQiwiApi.core.session.holder import AbstractSessionHolder
from glQiwiApi.ext.webhook_url import WebhookURL
from glQiwiApi.qiwi.settings import QiwiRouter, QiwiKassaRouter, QiwiApiMethods
from glQiwiApi.types import (
    QiwiAccountInfo,
    Transaction,
    Statistic,
    Limit,
    Account,
    Balance,
    Identification,
    CurrencyAmount,
    Card,
    Restriction,
    Commission,
    CrossRate,
    PaymentMethod,
    FreePaymentDetailsFields,
    PaymentInfo,
    OrderDetails,
    Bill,
    PlainAmount,
    P2PKeys,
    RefundBill,
    WebHookConfig,
    N,
    TransactionType,
    InvoiceStatus,
)
from glQiwiApi.types.arbitrary.file import File
from glQiwiApi.types.arbitrary.inputs import BinaryIOInput
from glQiwiApi.types.errors import QiwiErrorAnswer
from glQiwiApi.types.qiwi.transaction import Source
from glQiwiApi.utils.compat import Final
from glQiwiApi.utils.dates_conversion import \
    datetime_to_iso8601_with_moscow_timezone
from glQiwiApi.utils.exceptions import APIError, InvalidPayload, \
    ChequeIsNotAvailable
from glQiwiApi.utils.helper import allow_response_code, \
    override_error_messages, require
from glQiwiApi.utils.payload import (
    make_payload,
    format_dates,
    parse_iterable_to_list_of_objects,
    patch_p2p_create_payload,
    get_new_card_data,
    get_qiwi_master_data,
    parse_commission_request_payload,
    retrieve_card_data,
    set_data_to_wallet,
    check_dates_for_statistic_request,
    parse_limits,
    check_transaction,
    filter_none,
)
from glQiwiApi.utils.validators import PhoneNumber, String

MAX_HISTORY_TRANSACTION_LIMIT: Final[int] = 50
AmountType = Union[int, float]


def _is_copy_signal(kwargs: Dict[Any, bool]) -> bool:
    try:
        return kwargs.pop("__copy_signal__")
    except KeyError:
        return False


class QiwiWrapper(
    Wrapper,
    HandlerCollectionMixin,
    DataMixin,
    ContextInstanceMixin["QiwiWrapper"]
):
    """
    Delegates the work of QIWI API, webhooks, polling.
    Fast and versatile wrapper.

    """

    # declarative validators for fields
    phone_number = PhoneNumber(maxsize=15, minsize=11, optional=True)
    api_access_token = String(optional=True)
    secret_p2p = String(optional=True)

    def __init__(
            self,
            api_access_token: Optional[str] = None,
            phone_number: Optional[str] = None,
            secret_p2p: Optional[str] = None,
            cache_time_in_seconds: Union[float, int] = NO_CACHING,  # zero by default
            session_holder: Optional[AbstractSessionHolder[Any]] = None,
    ) -> None:
        """
        :param api_access_token: QIWI API token received from https://qiwi.com/api
        :param phone_number: your phone number starting with +
        :param secret_p2p: QIWI P2P secret key received from https://p2p.qiwi.com/
        :param cache_time_in_seconds: Time to cache requests in seconds,
                           default 0, respectively the request will not use the cache by default
        :param session_holder: obtains session and helps to manage session lifecycle. You can pass
                               your own session holder, for example using httpx lib and use it
        """
        self.phone_number = phone_number
        self._router = QiwiRouter()
        self._p2p_router = QiwiKassaRouter()
        self._request_service = RequestService(
            error_messages=self._router.config.ERROR_CODE_MESSAGES,
            cache_time=cache_time_in_seconds,
            session_holder=session_holder
        )
        self.api_access_token = api_access_token
        self.secret_p2p = secret_p2p

        self.set_current(self)

    def get_request_service(self) -> RequestService:
        return self._request_service

    def _add_authorization_header(self, headers: Dict[Any, Any],
                                  p2p: bool = False) -> Dict[Any, Any]:
        auth = headers.get("Authorization")
        if auth is None:
            auth = "Bearer {token}"
        if p2p:
            token = self.secret_p2p
        else:
            token = self.api_access_token
        headers["Authorization"] = auth.format(token=token)
        return headers

    def __new__(
            cls: Type[N],
            api_access_token: Optional[str] = None,
            phone_number: Optional[str] = None,
            secret_p2p: Optional[str] = None,
            without_context: bool = False,
            cache_time: Union[float, int] = NO_CACHING,
            *args: Any,
            **kwargs: Any,
    ) -> N:
        if (
                not isinstance(api_access_token, str)
                and not isinstance(secret_p2p, str)  # noqa: W503
                and not _is_copy_signal(kwargs)  # noqa: W503
        ):
            raise RuntimeError("Unable to initialize instance without tokens")

        return super().__new__(cls)  # type: ignore

    async def _register_webhook(
            self, web_url: Optional[str], txn_type: int = 2
    ) -> WebHookConfig:
        """
        This method register a new webhook

        :param web_url: service url
        :param txn_type:  0 => incoming, 1 => outgoing, 2 => all
        :return: Active Hooks
        """
        response = await self._request_service.api_request(
            "PUT",
            QiwiApiMethods.REG_WEBHOOK,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()
            ),
            params={"hookType": 1, "param": web_url, "txnType": txn_type},
        )
        return WebHookConfig.parse_obj(response)

    async def get_current_webhook(self) -> WebHookConfig:
        """
        List of active (active) notification handlers, associated with your wallet can be obtained with this request.
        Since now only one type of hook is used - webhook, then the response contains only one data object
        """
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.GET_CURRENT_WEBHOOK,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()
            ),
        )
        return WebHookConfig.parse_obj(response)

    async def _send_test_notification(self) -> Dict[Any, Any]:
        """
        Use this request to test your webhooks handler.
        Test notification is sent to the address specified during the call register_webhook
        """
        return await self._request_service.api_request(
            "GET",
            QiwiApiMethods.SEND_TEST_NOTIFICATION,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()
            ),
        )

    async def get_webhook_secret_key(self, hook_id: str) -> str:
        """
        Each notification contains a digital signature of the message, encrypted with a key.
        To obtain a signature verification key, use this request.

        :param hook_id: UUID of webhook
        :return: Base64 encoded key
        """
        resp = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.GET_WEBHOOK_SECRET,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()),
            hook_id=hook_id,
        )
        return cast(str, resp["key"])

    async def delete_current_webhook(self) -> Optional[Dict[str, str]]:
        """Method to delete webhook"""
        try:
            hook = await self.get_current_webhook()
        except APIError as ex:
            raise APIError(
                message=" You didn't register any webhook to delete ",
                status_code="422",
                traceback_info=ex.traceback_info,
            ) from None
        return await self._request_service.api_request(
            "DELETE",
            QiwiApiMethods.DELETE_CURRENT_WEBHOOK,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()
            ),
            hook_id=hook.hook_id,
        )

    async def change_webhook_secret(self, hook_id: str) -> str:
        """
        Use this request to change the encryption key for notifications.

        :param hook_id: UUID of webhook
        :return: Base64 encoded key
        """
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.CHANGE_WEBHOOK_SECRET,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()
            ),
            hook_id=hook_id,
        )
        return cast(str, response["key"])

    async def bind_webhook(
            self,
            url: Optional[Union[str, WebhookURL]] = None,
            transactions_type: int = 2,
            *,
            send_test_notification: bool = False,
            delete_old: bool = False,
    ) -> Tuple[Optional[WebHookConfig], str]:
        """
        [NON-API] EXCLUSIVE method to register new webhook or get old

        :param url: service url
        :param transactions_type: 0 => incoming, 1 => outgoing, 2 => all
        :param send_test_notification:  test_qiwi will send
         you test webhook update
        :param delete_old: boolean, if True - delete old webhook

        :return: Tuple of Hook and Base64-encoded key
        """
        key: Optional[str] = None
        if isinstance(url, WebhookURL):
            url = url.render()

        if delete_old:
            with suppress(APIError):
                await self.delete_current_webhook()

        try:
            # Try to register new webhook
            webhook = await self._register_webhook(url, transactions_type)
        except (APIError, TypeError):
            # Catching exceptions, if webhook already was registered or TypeError because missing url to bind
            try:
                webhook = await self.get_current_webhook()
            except APIError as ex:
                raise APIError(
                    message="An error occurred while getting the current webhook config. Most likely you did not call "
                            "method `QiwiWrapper.bind_webhook(...)` to register webhook"
                            " or did not pass on url to register it",
                    status_code=ex.status_code,
                    traceback_info=ex.traceback_info,
                ) from None
            key = await self.get_webhook_secret_key(webhook.hook_id)
            return webhook, key

        if send_test_notification:
            await self._send_test_notification()

        if not isinstance(key, str):
            key = await self.get_webhook_secret_key(webhook.hook_id)

        return webhook, key

    @override_error_messages(
        {
            404: {
                "message": "Wrong card number entered, possibly"
                           "the card to which you transfer is blocked"
            }
        }
    )
    async def _detect_mobile_number(self, phone_number: str) -> str:
        """
        Help method for getting phone ID

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = self._router.generate_default_headers()
        headers.update({"Content-Type": "application/x-www-form-urlencoded"})
        response = await self._request_service.raw_request(
            url="https://qiwi.com/mobile/detect.action",
            method="POST",
            headers=headers,
            data={"phone": phone_number},
        )
        return cast(str, response["message"])

    async def get_balance(self, *, account_number: int = 1) -> CurrencyAmount:
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.GET_BALANCE,
            self._router,
            headers=headers,
            phone_number=self.phone_number,
        )
        return CurrencyAmount.parse_obj(
            response["accounts"][account_number - 1]["balance"]
        )

    async def transactions(
            self,
            rows: int = MAX_HISTORY_TRANSACTION_LIMIT,
            operation: TransactionType = TransactionType.ALL,
            sources: Optional[List[Source]] = None,
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None,
    ) -> List[Transaction]:
        """
        Method for receiving transactions on the account
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param rows: number of transactions you want to receive
        :param operation: The type of operations in the report for selection.
        :param sources: List of payment sources, for filter
        :param start_date: The starting date for searching for payments.
                            Used only in conjunction with end_date.
        :param end_date: the end date of the search for payments.
                            Used only in conjunction with start_date.
        """
        if rows > 50 or rows <= 0:
            raise InvalidPayload("You can check no more than 50 transactions")
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        payload_data = format_dates(
            start_date=start_date,
            end_date=end_date,
            payload_data={
                "rows": rows,
                "operation": operation.value
            },
        )
        if sources is not None:
            for index, source in enumerate(sources, start=1):
                payload_data.update({f"source[{index}]": source.value})
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.TRANSACTIONS,
            self._router,
            params=payload_data,
            headers=headers,
            stripped_number=self.phone_number,
        )
        return parse_iterable_to_list_of_objects(
            cast(Iterable[Any], response.get("data")), Transaction
        )

    async def transaction_info(
            self, transaction_id: Union[str, int],
            transaction_type: TransactionType
    ) -> Transaction:
        """
        Method for obtaining complete information about a transaction

        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#txn_info

        :param transaction_id:
        :param transaction_type: only IN or OUT
        :return: Transaction object
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        payload_data = {"type": transaction_type.value}
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.TRANSACTION_INFO,
            self._router,
            headers=headers,
            params=payload_data,
            transaction_id=transaction_id,
        )
        return Transaction.parse_obj(response)

    async def get_restriction(self) -> List[Restriction]:
        """
        Method to check limits on your qiwi wallet
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#restrictions

        :return: List where the dictionary is located with restrictions,
         if there are no restrictions, it returns an empty list
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.CHECK_RESTRICTION,
            self._router,
            headers=headers,
            phone_number=self.phone_number,
        )
        return parse_iterable_to_list_of_objects(iterable=response,
                                                 model=Restriction)

    async def get_identification(self) -> Identification:
        """
        This method allows get your wallet identification data
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#ident
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.GET_IDENTIFICATION,
            self._router,
            headers=headers,
            phone_number=self.phone_number,
        )
        return Identification.parse_obj(response)

    async def check_transaction(
            self,
            amount: AmountType,
            transaction_type: TransactionType = TransactionType.IN,
            sender: Optional[str] = None,
            rows_num: int = 50,
            comment: Optional[str] = None,
    ) -> bool:
        """
        [ NON API METHOD ]

        Method for verifying a transaction.
        This method uses self.transactions (rows = rows) "under the hood" to check payment.

        For a little optimization, you can decrease rows by setting it,
        however, this does not guarantee the correct result

        Possible values for the transaction_type parameter:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'


        :param amount: amount of payment
        :param transaction_type: type of payment
        :param sender: number of receiver
        :param rows_num: number of payments to be checked
        :param comment: comment by which the transaction will be verified
        """
        if rows_num > 50 or rows_num <= 0:
            raise InvalidPayload("You can check no more than 50 transactions.")
        transactions = await self.transactions(rows=rows_num)
        return check_transaction(
            transactions=transactions,
            transaction_type=transaction_type,
            comment=comment,
            amount=amount,
            sender=sender,
        )

    async def get_limits(
            self, limit_types: Optional[List[str]] = None
    ) -> Dict[str, Limit]:
        """
        Function for getting limits on the qiwi wallet account
        Returns wallet limits as a list,
        if there is no limit for a certain country, then it does not include it in the list
        Detailed documentation:

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#limits
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers()
        )
        limit_types = limit_types or self._router.config.DEFAULT_LIMIT_TYPES
        # payload of limit types must be dict in format like array of strings {
        #    "types[0]": "Some type",
        #    "types[1]": "some other type",
        #    "types[n]": "n type"
        # }
        payload = {
            f"types[{index}]": limit_type
            for index, limit_type in enumerate(limit_types)
        }
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.GET_LIMITS,
            self._router,
            headers=headers,
            params=payload,
            stripped_number=self.phone_number,
        )
        return parse_limits(response)

    async def get_list_of_cards(self) -> List[Card]:
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET", QiwiApiMethods.GET_LIST_OF_CARDS, self._router,
            headers=headers
        )
        return parse_iterable_to_list_of_objects(iterable=response, model=Card)

    async def authenticate(
            self,
            birth_date: str,
            first_name: str,
            last_name: str,
            patronymic: str,
            passport: str,
            oms: Optional[str] = None,
            inn: Optional[str] = None,
            snils: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """
        This request allows you to send data to identify your QIWI wallet.
        It is allowed to identify no more than 5 wallets per owner

        To identify the wallet, you must send your full name, passport series number and date of birth.
        If the data has been verified, then the response will display
        your TIN and simplified wallet identification will be installed.
        If the data has not been verified,
        the wallet remains in the "Minimum" status.

        :param birth_date: Date of birth as a format string 1998-02-11
        :param first_name: First name
        :param last_name: Last name
        :param patronymic: Middle name
        :param passport: Series / Number of the passport. Ex: 4400111222
        :param oms:
        :param snils:
        :param inn:
        """

        payload = {
            "birthDate": birth_date,
            "firstName": first_name,
            "inn": inn,
            "lastName": last_name,
            "middleName": patronymic,
            "oms": oms,
            "passport": passport,
            "snils": snils,
        }
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        return await self._request_service.api_request(
            "POST",
            QiwiApiMethods.AUTHENTICATE,
            self._router,
            stripped_number=self.phone_number,
            headers=headers,
            data=filter_none(payload),
        )

    @override_error_messages(
        {
            422: {
                "message": "It is impossible to receive a check due to the fact that "
                           "the transaction for this ID has not been completed,"
                           "that is, an error occurred during the transaction"
            }
        }
    )
    async def get_receipt(
            self,
            transaction_id: Union[str, int],
            transaction_type: TransactionType,
            file_format: str = "PDF",
    ) -> File:
        """
        Method for receiving a receipt in byte format or file. \n
        Possible transaction_type values:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'

        :param transaction_id: transaction id, can be obtained by calling the to_wallet method, to_card
        :param transaction_type: type of transaction: 'IN', 'OUT', 'QIWI_CARD'
        :param file_format: format of file
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        data = {"type": transaction_type.value, "format": file_format}
        url = self._router.build_url(QiwiApiMethods.GET_RECEIPT,
                                     transaction_id=transaction_id)
        byte_response = await self._request_service.retrieve_bytes(
            url, "GET", params=data, headers=headers
        )
        try:
            err_model = QiwiErrorAnswer.parse_raw(byte_response)
            raise ChequeIsNotAvailable(err_model)
        except ValidationError:
            return File(BinaryIOInput.from_bytes(byte_response))

    async def get_account_info(self) -> QiwiAccountInfo:
        """
        Метод для получения информации об аккаунте

        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET", QiwiApiMethods.ACCOUNT_INFO, self._router, headers=headers
        )
        return QiwiAccountInfo.parse_obj(response)

    async def fetch_statistics(
            self,
            start_date: datetime,
            end_date: datetime,
            operation: TransactionType = TransactionType.ALL,
            sources: Optional[List[str]] = None,
    ) -> Statistic:
        """
        This query is used to get summary statistics
        by the amount of payments for a given period.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param start_date:The start date of the statistics period.
        :param end_date: End date of the statistics period.
        :param operation: The type of operations taken into account when calculating statistics.
         Allowed values:
            ALL - все операции,
            IN - только пополнения,
            OUT - только платежи,
            QIWI_CARD - только платежи по картам QIWI (QVC, QVP).
            По умолчанию ALL.
        :param sources: The sources of payments
            QW_RUB - рублевый счет кошелька,
            QW_USD - счет кошелька в долларах,
            QW_EUR - счет кошелька в евро,
            CARD - привязанные и непривязанные к кошельку банковские карты,
            MK - счет мобильного оператора. Если не указан,
            учитываются все источники платежа.
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        check_dates_for_statistic_request(start_date=start_date,
                                          end_date=end_date)
        params = {
            "startDate": datetime_to_iso8601_with_moscow_timezone(start_date),
            "endDate": datetime_to_iso8601_with_moscow_timezone(end_date),
            "operation": operation.value,
        }
        if sources:
            params.update({"sources": " ".join(sources)})
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.FETCH_STATISTICS,
            self._router,
            params=params,
            headers=headers,
            stripped_number=self.phone_number,
        )
        return Statistic.parse_obj(response)

    async def list_of_balances(self) -> List[Account]:
        """
        The request gets the current account balances of your QIWI Wallet.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#balances_list

        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.LIST_OF_BALANCES,
            self._router,
            headers=headers,
            stripped_number=self.phone_number,
        )
        return parse_iterable_to_list_of_objects(
            cast(Iterable[Any], response.get("accounts")), Account
        )

    @allow_response_code(201)
    async def create_new_balance(self, currency_alias: str) -> Optional[Dict[str, bool]]:
        """
        The request creates a new account and balance in your QIWI Wallet

        :param currency_alias: New account alias
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        payload = {"alias": currency_alias}
        return await self._request_service.api_request(
            "POST",
            QiwiApiMethods.CREATE_NEW_BALANCE,
            self._router,
            headers=headers,
            data=payload,
            stripped_number=self.phone_number,
        )

    async def available_balances(self) -> List[Balance]:
        """
        The request displays account aliases, available for creation in your QIWI Wallet

        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.AVAILABLE_BALANCES,
            self._router,
            headers=headers,
            stripped_number=self.phone_number,
        )
        return parse_iterable_to_list_of_objects(iterable=response,
                                                 model=Balance)

    @allow_response_code(status_code=204)
    async def set_default_balance(self, currency_alias: str) -> Dict[Any, Any]:
        """
        The request sets up an account for your QIWI Wallet, whose balance will be used for funding
        all payments by default.
        The account must be contained in the list of accounts, you can get the list by calling
        list_of_balances method

        :param currency_alias:
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        return await self._request_service.api_request(
            "PATCH",
            QiwiApiMethods.SET_DEFAULT_BALANCE,
            self._router,
            headers=headers,
            json={"defaultAccount": True},
            stripped_number=self.phone_number,
            currency_alias=currency_alias,
        )

    async def to_wallet(
            self,
            to_number: str,
            amount: Union[AmountType, str],
            currency: str = "643",
            comment: Optional[str] = None,
    ) -> PaymentInfo:
        """
        Method for transferring money to another wallet \n
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#p2p

        :param to_number: recipient number
        :param amount: the amount of money you want to transfer
        :param currency: special currency code
        :param comment: payment comment
        """
        data = set_data_to_wallet(
            data=deepcopy(self._router.config.QIWI_TO_WALLET),
            to_number=to_number,
            trans_sum=amount,
            currency=currency,
            comment=comment,
        )
        data.headers = self._add_authorization_header(headers=data.headers)
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.TO_WALLET,
            self._router,
            json=data.json,
            headers=data.headers,
        )
        return PaymentInfo.parse_obj(response)

    async def to_card(self, trans_sum: Union[float, int],
                      to_card: str) -> PaymentInfo:
        """
        Method for sending funds to the card.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#cards
        """
        data = retrieve_card_data(
            default_data=self._router.config.QIWI_TO_CARD,
            to_card=to_card,
            trans_sum=trans_sum,
            auth_maker=self._add_authorization_header,
        )
        privat_card_id = await self._detect_card_number(card_number=to_card)
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.TO_CARD,
            self._router,
            headers=data.headers,
            json=data.json,
            privat_card_id=privat_card_id,
        )
        return PaymentInfo.parse_obj(response)

    async def _detect_card_number(self, card_number: str) -> str:
        """
        Method for getting card ID

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        headers.update({"Content-Type": "application/x-www-form-urlencoded"})
        response = await self._request_service.raw_request(
            "https://qiwi.com/card/detect.action",
            "POST",
            headers=headers,
            data={"cardNumber": card_number},
        )
        return cast(str, response["message"])

    async def predict_commission(self, to_account: str,
                                 pay_sum: AmountType) -> Commission:
        """
        Full calc_commission of QIWI Wallet is refunded for payment in favor of the specified provider
        taking into account all tariffs for a given set of payment details.

        :param to_account:
        :param pay_sum:
        :return: Commission object
        """
        payload, special_code = parse_commission_request_payload(
            default_data=self._router.config.COMMISSION_DATA,
            auth_maker=self._add_authorization_header,
            to_account=to_account,
            pay_sum=pay_sum,
        )
        code: str = special_code or await self._detect_card_number(to_account)
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.COMMISSION,
            self._router,
            headers=payload.headers,
            json=payload.json,
            special_code=code,
        )
        return Commission.parse_obj(response)

    async def get_cross_rates(self) -> List[CrossRate]:
        """
        The method returns the current exchange rates and cross-rates of the QIWI Bank's currencies.

        """
        response = await self._request_service.api_request(
            "GET", QiwiApiMethods.GET_CROSS_RATES, self._router
        )
        return parse_iterable_to_list_of_objects(iterable=response["result"],
                                                 model=CrossRate)

    async def payment_by_payment_details(
            self,
            payment_sum: CurrencyAmount,
            payment_method: PaymentMethod,
            fields: FreePaymentDetailsFields,
            payment_id: Optional[str] = None,
    ) -> PaymentInfo:
        """
        Payment for services of commercial organizations according to their bank details.

        :param payment_id: payment id, if not transmitted, is used uuid4 by default
        :param payment_sum: a Sum object, which indicates the amount of the payment
        :param payment_method: payment method
        :param fields: payment details
        """
        payload = {
            "id": payment_id if isinstance(payment_id, str) else str(
                uuid.uuid4()),
            "sum": payment_sum.dict(),
            "paymentMethod": payment_method.dict(),
            "fields": fields.dict(),
        }
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.SPECIAL_PAYMENT,
            self._router,
            headers=self._router.generate_default_headers(),
            json=payload,
        )
        return PaymentInfo.parse_obj(response)

    async def buy_qiwi_master(self) -> PaymentInfo:
        """
        Method for buying QIWI Master package
        To call API methods, you need the QIWI Wallet API token with permissions to do the following:
        1. Management of virtual cards,
        2. Request information about the wallet profile,
        3. View payment history,
        4. Making payments without SMS.
        You can choose these rights when creating a new api token, to use api QIWI Master
        """
        payload = get_qiwi_master_data(
            cast(str, self.phone_number),
            self._router.config.QIWI_MASTER
        )
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.BUY_QIWI_MASTER,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()),
            json=payload,
        )
        return PaymentInfo.parse_obj(response)

    async def _pre_qiwi_master_request(self, card_alias: str = "qvc-cpa") -> OrderDetails:
        """Method for Issuing QIWI Master Virtual Card"""
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.PRE_QIWI_REQUEST,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()
            ),
            json={"cardAlias": card_alias},
            stripped_number=self.phone_number,
        )
        return OrderDetails.parse_obj(response)

    async def _confirm_qiwi_master_request(self,
                                           card_alias: str = "qvc-cpa") -> OrderDetails:
        """Confirmation of the card issue order"""
        details = await self._pre_qiwi_master_request(card_alias)
        response = await self._request_service.api_request(
            "PUT",
            QiwiApiMethods.CONFIRM_QIWI_MASTER,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()),
            stripped_number=self.phone_number,
            order_id=details.order_id,
        )
        return OrderDetails.parse_obj(response)

    async def _buy_new_qiwi_card(self, **kwargs: Any) -> Optional[OrderDetails]:
        kwargs.update(data=self._router.config.QIWI_MASTER)
        payload = get_new_card_data(**kwargs)
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.BUY_QIWI_CARD,
            self._router,
            json=payload,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()),
        )
        return OrderDetails.parse_obj(response)

    async def issue_qiwi_master_card(
            self, card_alias: str = "qvc-cpa"
    ) -> Optional[OrderDetails]:
        """
        Issuing a new card using the Qiwi Master API

        When issuing a card, 3, and possibly 3 requests are made, namely,
        according to the following scheme:
            - _pre_qiwi_master_request - this method creates a request
            - _confirm_qiwi_master_request - confirms the issue of the card
            - _buy_new_qiwi_card - buys a new card,
              if such a card is not free
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#qiwi-master-issue-card
        """
        pre_response = await self._confirm_qiwi_master_request(card_alias)
        if pre_response.status == "COMPLETED":
            return pre_response
        return await self._buy_new_qiwi_card(
            ph_number=self.phone_number, order_id=pre_response.order_id
        )

    async def _cards_qiwi_master(self) -> Dict[Any, Any]:
        return await self._request_service.api_request(
            "GET",
            QiwiApiMethods.CARDS_QIWI_MASTER,
            self._router,
            headers=self._add_authorization_header(
                self._router.generate_default_headers()),
        )

    @require("secret_p2p")
    async def reject_p2p_bill(self, bill_id: str) -> Bill:
        """Use this method to cancel unpaid invoice."""
        data = deepcopy(self._p2p_router.config.P2P_DATA)
        headers = self._add_authorization_header(data.headers, p2p=True)
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.REJECT_P2P_BILL,
            self._p2p_router,
            headers=headers,
            bill_id=bill_id,
        )
        return Bill.parse_obj(response)

    @require("secret_p2p")
    async def check_p2p_bill_status(self, bill_id: str) -> str:
        """
        Method for checking the status of a p2p transaction.\n
        Possible transaction types: \n
        WAITING	Bill is waiting for pay	\n
        PAID	Bill was paid	\n
        REJECTED	Bill was rejected\n
        EXPIRED	The bill has expired. Invoice not paid\n
        Docs:
        https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id:
        :return: status of bill
        """
        data = deepcopy(self._router.config.P2P_DATA)
        headers = self._add_authorization_header(data.headers, p2p=True)
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.CHECK_P2P_BILL_STATUS,
            self._p2p_router,
            headers=headers,
            bill_id=bill_id,
        )
        return Bill.parse_obj(response).status.value

    @require("secret_p2p")
    async def create_p2p_bill(
            self,
            amount: AmountType,
            bill_id: Optional[str] = None,
            comment: Optional[str] = None,
            life_time: Optional[datetime] = None,
            theme_code: Optional[str] = None,
            pay_source_filter: Optional[List[str]] = None,
    ) -> Bill:
        """
        It is the reliable method for integration.
        Parameters are sent by means of server2server requests with authorization.
        Method allows you to issue an invoice, successful
         response contains payUrl link to redirect client on Payment Form.
        Possible values of pay_source_filter:
          - 'qw'
          - 'card'
          - 'mobile'

        :param amount: amount of payment
        :param bill_id: unique transaction number, if not transmitted,
         generated automatically,
        :param life_time: the date until which the invoice will be available for payment.
        :param comment:
        :param theme_code:
        :param pay_source_filter: When you open the form, the following will be displayed
         only the translation methods specified in this parameter
        """
        if not isinstance(bill_id, (str, int)):
            bill_id = str(uuid.uuid4())
        life_time = datetime_to_iso8601_with_moscow_timezone(  # type: ignore
            life_time or constants.get_default_bill_time())
        data = deepcopy(self._p2p_router.config.P2P_DATA)
        headers = self._add_authorization_header(data.headers, p2p=True)
        payload = patch_p2p_create_payload(
            data, amount, str(life_time), comment, theme_code,
            pay_source_filter
        )
        response = await self._request_service.api_request(
            "PUT",
            QiwiApiMethods.CREATE_P2P_BILL,
            self._p2p_router,
            headers=headers,
            json=payload,
            bill_id=bill_id,
        )
        return Bill.parse_obj(response)

    async def retrieve_bills(
            self, rows: int, statuses: str = constants.DEFAULT_BILL_STATUSES
    ) -> List[Bill]:
        """
        A method for getting a list of your wallet's outstanding bills.

        The list is built in reverse chronological order.

        By default, the list is paginated with 50 items each,
        but you can specify a different number of elements (no more than 50).

        Filters by billing time can be used in the request,
        the initial account identifier.
        """
        params = make_payload(**locals())
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        if rows > 50 or rows <= 0:
            raise InvalidPayload("You can get no more than 50 invoices")
        response = await self._request_service.api_request(
            "GET",
            QiwiApiMethods.GET_BILLS,
            self._router,
            headers=headers,
            params=params,
        )
        return parse_iterable_to_list_of_objects(response["bills"], Bill)

    async def pay_the_invoice(self, invoice_uid: str,
                              currency: str) -> InvoiceStatus:
        """
        Execution of unconditional payment of the invoice without SMS-confirmation.

        ! Warning !
        To use this method correctly you need to tick "Проведение платежей без SMS"
        when registering QIWI API and retrieve token

        :param invoice_uid: Bill ID in QIWI system
        :param currency:
        """
        payload = make_payload(**locals())
        headers = self._add_authorization_header(
            self._router.generate_default_headers())
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.GET_BILLS,
            self._router,
            headers=headers,
            json=payload,
        )
        return InvoiceStatus.parse_obj(response)

    async def refund_bill(
            self,
            bill_id: Union[str, int],
            refund_id: Union[str, int],
            json_bill_data: Union[PlainAmount, Dict[str, Union[str, int]]],
    ) -> RefundBill:
        """
        The method allows you to make a refund on a paid invoice.
         in the JSON body of the request for the json_bill_data parameter:
         amount.value - refund amount.
         amount.currency - return currency.
        Can be a dictionary or an OptionalSum object
         Dictionary example: {
        "amount": {
            "currency": "RUB",
            "value": 1
            }
        }

        :param bill_id: unique account identifier in the merchant's system
        :param refund_id: unique identifier of the refund in the merchant's system.
        :param json_bill_data:
        :return: RefundBill object
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers(), p2p=True
        )
        json = (
            json_bill_data
            if isinstance(json_bill_data, dict)
            else json_bill_data.json()
        )
        response = await self._request_service.api_request(
            "PUT",
            QiwiApiMethods.REFUND_BILL,
            self._router,
            headers=headers,
            json=json,
            refund_id=refund_id,
            bill_id=bill_id,
        )
        return RefundBill.parse_obj(response)

    async def create_p2p_keys(
            self, key_pair_name: str,
            server_notification_url: Optional[str] = None
    ) -> P2PKeys:
        """
        Creates new pair of P2P keys to interact with P2P QIWI API

        :param key_pair_name: P2P token pair name
        :param server_notification_url: url for webhooks
        """
        headers = self._add_authorization_header(
            self._router.generate_default_headers(), p2p=True
        )
        data = {
            "keysPairName": key_pair_name,
            "serverNotificationsUrl": server_notification_url,
        }
        response = await self._request_service.api_request(
            "POST",
            QiwiApiMethods.CREATE_P2P_KEYS,
            self._router,
            headers=headers,
            json=data,
        )
        return P2PKeys.parse_obj(response)



from typing import Union, Optional, Type, Any

from glQiwiApi.core.dispatcher.filters import BaseFilter
from glQiwiApi.types import Transaction, WebHook, Notification


class TransactionFilter(BaseFilter[Transaction]):
    async def check(self, update: Union[Transaction, WebHook]) -> bool:
        return isinstance(update, (WebHook, Transaction))


class BillFilter(BaseFilter[Notification]):
    async def check(self, update: Notification) -> bool:
        return isinstance(update, Notification)


class ErrorFilter(BaseFilter[Exception]):
    def __init__(self, exception: Optional[Union[Type[Exception], Exception]] = None):
        self.exception: Union[Type[Exception], Exception] = exception or Exception

    async def check(self, error: Union[Type[Exception], Exception], *args: Any) -> bool:
        try:
            raise error
        except self.exception:  # type: ignore
            return True
        except:  # noqa
            raise error



import typing

from aiohttp import web

from glQiwiApi.core.dispatcher.implementation import Dispatcher
from glQiwiApi.core.dispatcher.webhooks.config import (
    DEFAULT_QIWI_WEBHOOK_PATH,
    DEFAULT_QIWI_ROUTER_NAME,
    DEFAULT_QIWI_BILLS_ROUTER_NAME,
    DEFAULT_QIWI_BILLS_WEBHOOK_PATH,
    Path,
)
from glQiwiApi.core.dispatcher.webhooks.utils import check_ip
from glQiwiApi.core.dispatcher.webhooks.views.bill_view import QiwiBillWebView
from glQiwiApi.core.dispatcher.webhooks.views.transaction_view import QiwiWebHookWebView


def configure_transaction_view(
        app: web.Application,
        dispatcher: Dispatcher,
        path: typing.Optional[Path] = None,
) -> None:
    app[QiwiWebHookWebView.app_key_check_ip] = check_ip
    app[QiwiWebHookWebView.app_key_dispatcher] = dispatcher
    if isinstance(path, Path):
        txn_path = path.transaction_path
    else:
        txn_path = DEFAULT_QIWI_WEBHOOK_PATH
    app.router.add_view(
        typing.cast(str, txn_path), QiwiWebHookWebView, name=DEFAULT_QIWI_ROUTER_NAME
    )


def configure_bill_view(
        app: web.Application,
        secret_key: typing.Optional[str],
        dispatcher: Dispatcher,
        path: typing.Optional[Path] = None,
) -> None:
    app["_secret_key"] = secret_key
    app[QiwiBillWebView.app_key_check_ip] = check_ip
    app[QiwiBillWebView.app_key_dispatcher] = dispatcher
    if isinstance(path, Path):
        bill_path = path.bill_path
    else:
        bill_path = DEFAULT_QIWI_BILLS_WEBHOOK_PATH
    app.router.add_view(
        handler=QiwiBillWebView,
        name=DEFAULT_QIWI_BILLS_ROUTER_NAME,
        path=typing.cast(str, bill_path),
    )


def configure_app(
        dispatcher: Dispatcher,
        app: web.Application,
        path: typing.Optional[Path] = None,
        secret_key: typing.Optional[str] = None
) -> web.Application:
    """
    Entirely configures the web app for webhooks

    :param dispatcher: dispatcher, which processing events
    :param app: aiohttp.web.Application
    :param path: Path obj, contains two paths
    :param secret_key: secret p2p key
    """

    configure_bill_view(app, secret_key, dispatcher, path)
    configure_transaction_view(app, dispatcher, path)
    return app



from __future__ import annotations

import abc

from glQiwiApi.types import Transaction, CurrencyAmount
from .base import Handler, ClientMixin


class AbstractTransactionHandler(
    Handler[Transaction], ClientMixin[Transaction], abc.ABC
):
    @property
    def transaction_id(self) -> int:
        return self.event.id

    @property
    def transaction_sum(self) -> CurrencyAmount:
        return self.event.sum



from typing import Tuple, Any, Dict, ItemsView


class Polygon:
    """Polygon class for QiwiMaps class"""

    def __init__(
        self, lat_lon_pair_nw: Tuple[Any, ...], lat_lon_pair_se: Tuple[Any, ...]
    ) -> None:
        self.lat_nw, self.lon_nw = lat_lon_pair_nw
        self.lat_se, self.lon_se = lat_lon_pair_se

        self._dict = {
            "latNW": self.lat_nw,
            "lngNW": self.lon_nw,
            "latSE": self.lat_se,
            "lngSE": self.lon_se,
        }

    @property
    def dict(self) -> Dict[Any, Any]:
        return {k: str(double) for k, double in self._get_items()}

    def _get_items(self) -> ItemsView[str, Any]:
        return self._dict.items()


__all__ = ("Polygon",)



from __future__ import annotations

import abc
import logging
from typing import Optional, Generic, TypeVar, TYPE_CHECKING, Tuple, cast, Any, Type

from aiohttp import web
from pydantic import ValidationError

from glQiwiApi.core.dispatcher.implementation import Dispatcher
from glQiwiApi.core.dispatcher.webhooks.collision_detectors import HashBasedCollisionDetector, \
    UnexpectedCollision

if TYPE_CHECKING:  # pragma: no cover
    from glQiwiApi.types.base import HashableBase  # pragma: no cover  # noqa

try:
    import orjson as json
except ImportError:
    import json  # type: ignore

Event = TypeVar("Event", bound="HashableBase")

logger = logging.getLogger("glQiwiApi.webhooks.base")


class BaseWebHookView(web.View, Generic[Event]):
    """
    Base webhook view for processing events
    You can make own view and than use it in code,
    just inheriting this base class

    """

    app_key_check_ip: Optional[str] = None
    """app_key_check_ip stores key to a callable, which check ip"""

    app_key_dispatcher: Optional[str] = None
    """app_key_handler_dispatcher store Dispatcher instance to feed events to handlers"""

    collision_detector: HashBasedCollisionDetector[Event] = HashBasedCollisionDetector()
    """
    QIWI API can send the same update twice or more, so we need to avoid this problem anyway.
    Also, you can override it with redis usage or more advanced hashing.
    """

    _event_type: Type[Event]
    """Has to be redefined"""

    async def parse_update(self) -> Event:
        """Parse raw update and return pydantic model"""
        data = await self._json_body()
        try:
            if isinstance(data, str):
                return self._event_type.parse_raw(data)
            elif isinstance(data, dict):
                return self._event_type.parse_obj(data)
            else:
                raise ValidationError
        except ValidationError:
            raise web.HTTPBadRequest()

    async def _json_body(self) -> Any:
        text = await self.request.text()
        return json.loads(text)

    def check_ip(self) -> Tuple[Optional[str], bool]:
        """
        Check client IP. Accept requests only from telegram servers.

        :return:
        """
        if self.app_key_check_ip is None:
            raise RuntimeError(
                "There is no check_ip function to validate ip."
                " Please, override `app_key_check_ip` attribute"
            )

        check_ip_function = self.request.app.get(self.app_key_check_ip, False)

        # For reverse proxy (nginx)
        forwarded_for = self.request.headers.get("X-Forwarded-For")
        if forwarded_for:
            return forwarded_for, check_ip_function(forwarded_for)

        # For default method
        peer_name = self.request.transport.get_extra_info("peername")  # type: ignore
        if peer_name is not None:
            host, _ = peer_name
            return cast(Optional[str], host), check_ip_function(host)

        # Not allowed and can't get client IP
        return None, False

    def validate_ip(self) -> None:
        """
        Check ip if that is needed. Raise web.HTTPUnauthorized for not allowed hosts.
        """
        if self.request.app.get(self.app_key_check_ip, False):  # type: ignore
            ip_address, accept = self.check_ip()
            if not accept:
                logger.debug(f"Blocking request from an unauthorized IP: {ip_address}")
                raise web.HTTPUnauthorized()
        return None  # hint for mypy

    async def process_event(self, event: Event) -> None:
        await self.dispatcher.process_event(event)

    async def post(self) -> web.Response:
        """
        Process POST request with basic IP validation.

        """
        self.validate_ip()
        event = await self.parse_update()

        try:
            self.collision_detector.remember_processed_object(event)
        except UnexpectedCollision:
            logger.debug("Detect collision on event %s", event)
            return self.ok_response()

        self._validate_event_signature(event)
        await self.process_event(event)
        return self.ok_response()

    async def get(self) -> web.Response:
        self.validate_ip()
        return web.Response(text="")

    @abc.abstractmethod
    def ok_response(self) -> web.Response:
        ...

    @abc.abstractmethod
    def _validate_event_signature(self, update: Event) -> None:
        raise NotImplementedError

    @property
    def dispatcher(self) -> Dispatcher:
        if self.app_key_dispatcher is None:
            raise RuntimeError(
                "There is no check_ip function to validate ip."
                " Please, override `app_key_dispatcher` attribute"
            )
        return cast(Dispatcher, self.request.app[self.app_key_dispatcher])



from __future__ import annotations

import typing as t

from glQiwiApi.plugins.telegram.base import TelegramPlugin

if t.TYPE_CHECKING:
    from glQiwiApi.utils.compat import Dispatcher


class TelegramPollingPlugin(TelegramPlugin):
    """
    Builtin telegram proxy.
    Allows you to use Telegram and QIWI webhooks together

    """

    def __init__(
            self,
            dispatcher: Dispatcher,
            timeout: int = 20,
            relax: float = 0.1,
            limit: t.Optional[t.Any] = None,
            reset_webhook: t.Optional[t.Any] = None,
            fast: t.Optional[bool] = True,
            error_sleep: int = 5,
            allowed_updates: t.Optional[t.List[str]] = None,
    ) -> None:
        super(TelegramPollingPlugin, self).__init__(dispatcher)
        self._allowed_updates = allowed_updates
        self._error_sleep = error_sleep
        self._fast = fast
        self._reset_webhook = reset_webhook
        self._timeout = timeout
        self._relax = relax
        self._limit = limit

    async def incline(self, ctx: t.Dict[t.Any, t.Any]) -> t.Any:
        """
        Set up polling to run polling qiwi updates concurrently with aiogram

        :param ctx: you can pass on loop as key/value parameter
        """
        await self.dispatcher.start_polling(
            timeout=self._timeout,
            reset_webhook=self._reset_webhook,
            relax=self._relax,
            allowed_updates=self._allowed_updates,
            limit=self._limit,
            error_sleep=self._error_sleep,
            fast=self._fast,
        )

    async def shutdown(self) -> None:
        self.dispatcher.stop_polling()



from . import server  # noqa
from .base import BaseWebHookView
from .views import QiwiBillWebView, QiwiWebHookWebView

__all__ = ("QiwiBillWebView", "QiwiWebHookWebView", "BaseWebHookView", "server")



from __future__ import annotations

import logging
from typing import cast

from aiohttp import web

from glQiwiApi import types
from glQiwiApi.core.dispatcher.webhooks.base import BaseWebHookView
from glQiwiApi.types.exceptions import WebhookSignatureUnverified

logger = logging.getLogger("glQiwiApi.webhooks.p2p")


class QiwiBillWebView(BaseWebHookView[types.Notification]):
    """View, which processes p2p notifications"""
    _event_type = types.Notification

    def _validate_event_signature(self, update: types.Notification) -> None:
        sha256_signature = cast(str, self.request.headers.get("X-Api-Signature-SHA256"))
        webhook_base64 = cast(str, self.request.app.get("_secret_key"))

        try:
            update.verify_signature(sha256_signature, webhook_base64)
        except WebhookSignatureUnverified:
            logger.debug("Blocking request due to invalid signature of payload.")
            raise web.HTTPBadRequest()

    def ok_response(self) -> web.Response:
        return web.json_response(data={"error": "0"})

    app_key_check_ip = "_qiwi_bill_check_ip"
    app_key_dispatcher = "_qiwi_bill_dispatcher"



from __future__ import annotations

import logging
from typing import cast, Optional

from aiohttp import web

from glQiwiApi import types
from glQiwiApi.core.dispatcher.webhooks.base import BaseWebHookView
from glQiwiApi.types.exceptions import WebhookSignatureUnverified

logger = logging.getLogger("logging.webhooks.default")


class QiwiWebHookWebView(BaseWebHookView[types.WebHook]):
    """
    View, which processes transactions

    """
    _event_type = types.WebHook

    def _validate_event_signature(self, update: types.WebHook) -> None:
        base64_key = cast(Optional[str], self.request.app.get("_base64_key"))

        if update.is_testable:
            return None

        if not base64_key:
            logger.warning(
                "Validation was skipped because there is no base64 key to compare hash",
                UserWarning,
                stacklevel=2,
            )
            return None

        try:
            update.verify_signature(base64_key)
        except WebhookSignatureUnverified:
            logger.debug(
                "Request has being blocked due to invalid signature of json request payload."
            )
            raise web.HTTPBadRequest()

    def ok_response(self) -> web.Response:
        return web.Response(text="ok")

    app_key_check_ip = "_qiwi_wallet_check_ip"
    app_key_dispatcher = "_qiwi_wallet_dispatcher"



from .base import Handler
from .bill import AbstractBillHandler
from .error import ErrorHandler
from .transaction import AbstractTransactionHandler
from .webhook_transaction import AbstractWebHookHandler

__all__ = (
    "AbstractTransactionHandler",
    "AbstractBillHandler",
    "Handler",
    "AbstractWebHookHandler",
    "ErrorHandler",
)



from __future__ import annotations

import abc
import typing as t

from glQiwiApi.plugins.abc import Pluggable

if t.TYPE_CHECKING:
    from glQiwiApi.utils.compat import Dispatcher


class TelegramPlugin(Pluggable, abc.ABC):
    def __init__(self, dispatcher: Dispatcher) -> None:
        self.bot = dispatcher.bot
        self.dispatcher = dispatcher



from __future__ import annotations

from typing import TYPE_CHECKING, Hashable

from pydantic import BaseModel, Extra

if TYPE_CHECKING:
    from glQiwiApi.qiwi.client import QiwiWrapper


class Base(BaseModel):
    @property
    def client(self) -> "QiwiWrapper":
        """Returning an instance of :class:`QiwiWrapper`"""
        from glQiwiApi.qiwi.client import QiwiWrapper

        instance = QiwiWrapper.get_current()

        if instance is None:
            raise RuntimeError(
                "Can't get client instance from context. "
                "You can fix this by setting the current instance: "
                "`QiwiWrapper.set_current(...)`"
            )
        return instance


class HashableBase(Base):
    class Config:
        allow_mutation = False

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))

    def __eq__(self, other: Hashable) -> bool:
        return self.__hash__() == other.__hash__()


class ExtraBase(Base):
    class Config:  # pragma: no cover
        extra = Extra.allow



from __future__ import annotations

import typing

from glQiwiApi import types
from glQiwiApi.core import RequestService
from glQiwiApi.core.abc.wrapper import Wrapper
from glQiwiApi.core.constants import NO_CACHING
from glQiwiApi.core.mixins import ContextInstanceMixin, DataMixin
from glQiwiApi.core.session.holder import AbstractSessionHolder
from glQiwiApi.utils.payload import parse_iterable_to_list_of_objects, filter_none


class QiwiMaps(Wrapper, DataMixin, ContextInstanceMixin["QiwiMaps"]):
    """
    QIWI Terminal Maps API allows you to locate
    QIWI terminals on the territory of the Russian Federation

    """

    def __init__(self, cache_time: int = NO_CACHING,
                 session_holder: typing.Optional[AbstractSessionHolder[typing.Any]] = None) -> None:
        self._request_service = RequestService(cache_time=cache_time, session_holder=session_holder)

    def get_request_service(self) -> RequestService:
        return self._request_service

    async def terminals(
            self,
            polygon: types.Polygon,
            zoom: typing.Optional[int] = None,
            pop_if_inactive_x_mins: int = 30,
            include_partners: typing.Optional[bool] = None,
            partners_ids: typing.Optional[typing.List[typing.Any]] = None,
            cache_terminals: typing.Optional[bool] = None,
            card_terminals: typing.Optional[bool] = None,
            identification_types: typing.Optional[int] = None,
            terminal_groups: typing.Optional[typing.List[typing.Any]] = None,
    ) -> typing.List[types.Terminal]:
        """
        Get map of terminals sent for passed polygon with additional params

        :param polygon: glQiwiApi.types.Polygon object
        :param zoom:
         https://tech.yandex.ru/maps/doc/staticapi/1.x/dg/concepts/map_scale-docpage/
        :param pop_if_inactive_x_mins: do not show if terminal
         was inactive for X minutes default 0.5 hours
        :param include_partners: result will include/exclude partner terminals
        :param partners_ids: Not fixed IDS array look at docs
        :param cache_terminals: result will include or exclude
         cache-acceptable terminals
        :param card_terminals: result will include or
         exclude card-acceptable terminals
        :param identification_types: `0` - not identified, `1` -
         partly identified, `2` - fully identified
        :param terminal_groups: look at QiwiMaps.partners
        :return: list of Terminal instances
        """
        params = filter_none(
            {
                **polygon.dict,
                "zoom": zoom,
                "activeWithinMinutes": pop_if_inactive_x_mins,
                "withRefillWallet": include_partners,
                "ttpIds": partners_ids,
                "cacheAllowed": cache_terminals,
                "cardAllowed": card_terminals,
                "identificationTypes": identification_types,
                "ttpGroups": terminal_groups,
            }
        )
        url = "http://edge.qiwi.com/locator/v3/nearest/clusters?parameters"
        response = await self._request_service.raw_request(url, "GET", params=params)
        return parse_iterable_to_list_of_objects(
            typing.cast(typing.List[typing.Any], response), types.Terminal
        )

    async def partners(self) -> typing.List[types.Partner]:
        """
        Get terminal partners for ttpGroups
        :return: list of TTPGroups
        """
        url = "http://edge.qiwi.com/locator/v3/ttp-groups"
        response = await self._request_service.raw_request(
            url, "GET", headers={"Content-type": "text/json"}
        )
        return parse_iterable_to_list_of_objects(
            typing.cast(typing.List[typing.Any], response), types.Partner
        )



from __future__ import annotations

import abc
from typing import TypeVar, List, Generic

T = TypeVar("T")


class UnexpectedCollision(Exception):
    pass


class BaseCollisionDetector(abc.ABC, Generic[T]):

    @abc.abstractmethod
    def has_collision(self, obj: T) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def add_already_processed_event(self, obj: T) -> None:
        raise NotImplementedError

    def remember_processed_object(self, obj: T) -> None:
        if self.has_collision(obj):
            raise UnexpectedCollision()
        self.add_already_processed_event(obj)


class HashBasedCollisionDetector(BaseCollisionDetector[T]):

    def __init__(self) -> None:
        self.already_processed_object_hashes: List[int] = []

    def add_already_processed_event(self, obj: T) -> None:
        self.already_processed_object_hashes.append(hash(obj))

    def has_collision(self, obj: T) -> bool:
        return any(
            hash(obj) == processed_hash
            for processed_hash in self.already_processed_object_hashes
        )



from datetime import timedelta, datetime

DEFAULT_TIMEOUT = 5

NO_CACHING = 0

TIMEOUT_IF_EXCEPTION = 40


def get_default_bill_time() -> datetime:
    return datetime.now() + timedelta(days=2)


DEFAULT_BILL_STATUSES = "READY_FOR_PAY"

DEFAULT_APPLICATION_KEY = "__aiohttp_web_application__"



from .core import (
    BaseFilter,
    LambdaBasedFilter,
    Handler,
    AbstractBillHandler,
    AbstractTransactionHandler,
    AbstractWebHookHandler,
    async_as_sync,
    execute_async_as_sync,
)
from .qiwi import QiwiWrapper, QiwiMaps
from .types.exceptions import WebhookSignatureUnverified
from .utils.exceptions import (
    InvalidPayload,
    CantParseUrl,
    NetworkError,
    APIError,
    NoUpdatesToExecute,
    InvalidCachePayload,
    BadCallback,
    ChequeIsNotAvailable
)
from .yoo_money import YooMoneyAPI

try:
    import uvloop  # pragma: no cover  # NOQA

    uvloop.install()
except ImportError:
    pass  # pragma: no cover

__version__ = "1.1.4"

__all__ = (
    "QiwiWrapper",
    "YooMoneyAPI",
    "QiwiMaps",
    "execute_async_as_sync",
    "async_as_sync",
    "LambdaBasedFilter",
    "BaseFilter",
    # class-based handlers
    "AbstractBillHandler",
    "AbstractTransactionHandler",
    "AbstractWebHookHandler",
    "Handler",
    # Exceptions
    "InvalidPayload",
    "CantParseUrl",
    "APIError",
    "NoUpdatesToExecute",
    "NetworkError",
    "InvalidCachePayload",
    "BadCallback",
    "WebhookSignatureUnverified",
    "ChequeIsNotAvailable",
    # other
    "__version__",
)



from __future__ import annotations

import re
from copy import deepcopy
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import (
    Any,
    Union,
    Optional,
    Dict,
    Type,
    List,
    MutableMapping,
    Tuple,
    Iterable,
    cast,
    TypeVar,
)

from aiohttp import RequestInfo
from pydantic import BaseModel

from glQiwiApi import types
from glQiwiApi.utils.dates_conversion import datetime_to_iso8601_with_moscow_timezone, datetime_to_utc

try:
    import orjson
except (ModuleNotFoundError, ImportError):  # pragma: no cover # type: ignore
    import json as orjson  # type: ignore

from glQiwiApi.utils import exceptions

Model = TypeVar("Model", bound=BaseModel)
DEFAULT_EXCLUDE = ("cls", "self", "__class__")


def filter_none(dictionary: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Pop NoneType values and convert everything to str, designed?for=params

    :param dictionary: source dict
    :return: filtered dict
    """
    return {k: str(v) for k, v in dictionary.items() if v is not None}


def make_payload(**kwargs: Any) -> Dict[Any, Any]:
    return {
        key: value
        for key, value in kwargs.items()
        if key not in DEFAULT_EXCLUDE and value is not None
    }


def get_decoded_result(
        error_messages: Dict[int, str],
        status_code: int,
        request_info: RequestInfo,
        body: str,
) -> Dict[Any, Any]:
    """
    Checks whether `result` is a valid API response.
    A result is considered invalid if:
        - The server returned an HTTP response code other than 200

    :param error_messages:
    :param status_code: status code
    :param body: body of response
    :param request_info:
    :return: The result parsed to a JSON dictionary
    :raises ApiException: if one of the above listed cases is applicable
    """

    try:
        result_json = cast(Dict[Any, Any], orjson.loads(body))
    except ValueError:
        result_json = {}

    if status_code == HTTPStatus.OK:
        return result_json

    raise exceptions.APIError(
        message=error_messages.get(status_code),
        status_code=status_code,
        traceback_info=request_info,
    )


def parse_auth_link(response_data: str) -> str:
    """
    Parse link for getting code, which needs to be entered in the method
    get_access_token

    :param response_data:
    """
    regexp = re.compile(
        r"https://yoomoney.ru/oauth2/authorize[?]requestid[=]\w+"
    )  # pragma: no cover
    return cast(str, re.findall(regexp, str(response_data))[0])  # pragma: no cover


def format_dates(
        start_date: Optional[datetime],
        end_date: Optional[datetime],
        payload_data: Dict[Any, Any],
) -> Dict[Any, Any]:
    """Check correctness of transferred dates and add it to request"""
    if isinstance(start_date, datetime) and isinstance(end_date, datetime):
        if (end_date - start_date).total_seconds() > 0:
            payload_data.update(
                {
                    "startDate": datetime_to_iso8601_with_moscow_timezone(start_date),
                    "endDate": datetime_to_iso8601_with_moscow_timezone(end_date),
                }
            )
        else:
            raise ValueError("end_date cannot be bigger than start_date")
    return payload_data


def parse_commission_request_payload(
        default_data: types.WrappedRequestPayload,
        auth_maker: types.FuncT,
        pay_sum: Union[int, float],
        to_account: str,
) -> Tuple[types.WrappedRequestPayload, Union[str, None]]:
    """Set calc_commission payload"""
    payload = deepcopy(default_data)
    payload.headers = auth_maker(payload.headers)
    payload.json["purchaseTotals"]["total"]["amount"] = pay_sum
    payload.json["account"] = to_account
    return payload, "99" if len(to_account) <= 15 else None


def retrieve_card_data(
        default_data: types.WrappedRequestPayload,
        trans_sum: Union[int, float, str],
        to_card: str,
        auth_maker: types.FuncT,
) -> types.WrappedRequestPayload:
    """Set card data payload"""
    data = deepcopy(default_data)
    data.json["sum"]["amount"] = trans_sum
    data.json["fields"]["account"] = to_card
    data.headers = auth_maker(headers=data.headers)
    return data


def retrieve_base_headers_for_yoomoney(
        content_json: bool = False, auth: bool = False
) -> Dict[Any, Any]:
    headers = {
        "Host": "yoomoney.ru",
        "Content-Type": "application/x-www-form-urlencoded",
    }
    if content_json:
        headers.update({"Accept": "application/json"})
    if auth:
        headers.update({"Authorization": "Bearer {token}"})
    return headers


def set_data_to_wallet(
        data: types.WrappedRequestPayload,
        to_number: str,
        trans_sum: Union[str, int, float],
        comment: Optional[str] = None,
        currency: str = "643",
) -> types.WrappedRequestPayload:
    data.json["sum"]["amount"] = str(trans_sum)
    data.json["sum"]["currency"] = currency
    data.json["fields"]["account"] = to_number
    if comment is not None:
        data.json["comment"] = comment
    data.headers.update({"User-Agent": "Android v3.2.0 MKT"})
    return data


def patch_p2p_create_payload(
        wrapped_data: types.WrappedRequestPayload,
        amount: Union[str, int, float],
        life_time: str,
        comment: Optional[str] = None,
        theme_code: Optional[str] = None,
        pay_source_filter: Optional[List[str]] = None,
) -> Dict[MutableMapping[Any, Any], Any]:
    """Setting data for p2p form creation transfer"""
    wrapped_data.json["amount"]["value"] = str(amount)
    wrapped_data.json["comment"] = comment
    wrapped_data.json["expirationDateTime"] = life_time
    if isinstance(pay_source_filter, list):
        source_filters = " ".join(pay_source_filter).replace(" ", ",")
        wrapped_data.json["customFields"]["paySourcesFilter"] = source_filters
    if isinstance(theme_code, str):
        wrapped_data.json["customFields"]["themeCode"] = theme_code
    if not isinstance(theme_code, str) and pay_source_filter not in [  # type: ignore
        "qw",
        "card",
        "mobile",
    ]:
        wrapped_data.json.pop("customFields")
    return wrapped_data.json


def parse_iterable_to_list_of_objects(
        iterable: Iterable[Any], model: Type[Model]
) -> List[Model]:
    """
    Parse simple objects, which cant raise ValidationError

    :param iterable:
    :param model: pydantic model, which will parse data
    """
    return [model.parse_obj(obj) for obj in iterable]


def get_qiwi_master_data(ph_number: str, data: Dict[Any, Any]) -> Dict[Any, Any]:
    payload = deepcopy(data)
    payload["fields"]["account"] = ph_number
    return payload


def get_new_card_data(
        ph_number: str, order_id: str, data: Dict[Any, Any]
) -> Dict[Any, Any]:
    payload = deepcopy(data)
    payload["fields"].pop("vas_alias")
    payload["fields"].update(order_id=order_id)
    payload["fields"]["account"] = ph_number
    return payload


def parse_amount(
        txn_type: str, txn: types.OperationDetails
) -> Tuple[Union[int, float], str]:
    transaction_type_in_lower = types.OperationType.DEPOSITION.value.lower()  # type: str
    if txn_type == transaction_type_in_lower:
        return txn.amount, txn.comment  # type: ignore
    else:
        return txn.amount_due, txn.message  # type: ignore


def check_params(
        amount_: Union[int, float],
        amount: Union[int, float],
        txn: types.OperationDetails,
        transaction_type: str,
) -> bool:
    return amount is not None and amount <= amount_ and txn.direction == transaction_type


def check_transaction(
        transactions: List[types.Transaction],
        amount: Union[int, float],
        transaction_type: types.TransactionType = types.TransactionType.IN,
        sender: Optional[str] = None,
        comment: Optional[str] = None,
) -> bool:
    for txn in transactions:
        if txn.sum.amount < amount or txn.type != transaction_type.value:
            continue
        if txn.comment == comment and txn.to_account == sender:
            return True
        elif comment and sender:
            continue
        elif txn.to_account == sender:
            return True
        elif sender:
            continue
        elif txn.comment == comment:
            return True
        elif comment:
            continue
        return True
    return False


def parse_limits(response: Dict[Any, Any]) -> Dict[str, types.Limit]:
    resp_limits = cast(Dict[str, List[Dict[Any, Any]]], response["limits"])
    return {
        code: [types.Limit.parse_obj(limit) for limit in limits]  # type: ignore
        for code, limits in resp_limits.items()
    }


def check_api_method(api_method: str) -> None:
    if not isinstance(api_method, str):
        raise RuntimeError(
            f"Invalid type of api_method(must  be string)."
            f" Passed {type(api_method)}"
        )


def format_transactions_payload(
        data: Dict[Any, Any],
        records: int,
        operation_types: Optional[
            Union[List[types.OperationType], Tuple[types.OperationType, ...]]
        ] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        start_record: Optional[int] = None,
) -> Dict[Any, Any]:
    from glQiwiApi import InvalidPayload

    if records <= 0 or records > 100:
        raise InvalidPayload(
            "Invalid number of records."
            "The number of records that can be requested,"
            "be in the range from 1 to 100 inclusive"
        )
    if operation_types and all(
            isinstance(operation_type, types.OperationType)
            for operation_type in operation_types
    ):
        op_types = [operation_type.value for operation_type in operation_types]
        data.update({"type": " ".join(op_types)})

    if isinstance(start_record, int):
        if start_record < 0:
            raise InvalidPayload("start_record must be positive")
        data.update({"start_record": start_record})

    if start_date:
        if not isinstance(start_date, datetime):
            raise InvalidPayload("start_date must be datetime instance")
        data.update({"from": datetime_to_utc(start_date)})

    if end_date:
        if not isinstance(end_date, datetime):
            raise InvalidPayload("end_date must be datetime instance")
        data.update({"till": datetime_to_utc(end_date)})
    return data


def check_dates_for_statistic_request(start_date: datetime, end_date: datetime) -> None:
    first_expression = isinstance(start_date, datetime) and isinstance(
        end_date, datetime
    )
    second_expression = isinstance(start_date, timedelta) and isinstance(
        end_date, datetime
    )
    if first_expression or second_expression:
        delta: timedelta = end_date - start_date
        if delta.days > 90:
            raise ValueError(
                "The maximum period for downloading statistics is 90 calendar days."
            )
    else:
        raise ValueError(
            "You passed in the start and end date values in the wrong format."
        )



from __future__ import annotations

import asyncio
import functools
import logging
import operator
import types
from typing import (
    List,
    Tuple,
    Callable,
    Union,
    Awaitable,
    Optional,
    cast,
    Generic,
    TYPE_CHECKING,
    TypeVar,
    Type,
    Any,
    Iterable,
    no_type_check,
)

from glQiwiApi.core.dispatcher._builtin_filters import TransactionFilter, BillFilter, ErrorFilter
from .class_based import (
    AbstractTransactionHandler,
    AbstractBillHandler,
    Handler,
    ErrorHandler,
)
from .filters import BaseFilter, LambdaBasedFilter

if TYPE_CHECKING:
    from glQiwiApi.types import Notification, WebHook, Transaction  # pragma: no cover
    from glQiwiApi.types.base import HashableBase, Base  # noqa  # pragma: no cover

Event = TypeVar("Event", bound=Union["HashableBase", Exception, "Base"])
_T = TypeVar("_T")
EventFilter = Callable[[Event], bool]
TxnFilters = Union[
    BaseFilter["Transaction"],
    BaseFilter["WebHook"],
    Callable[["WebHook"], bool],
    Callable[["Transaction"], bool],
]

BillFilters = Union[Callable[["Notification"], bool], BaseFilter["Notification"]]

HandlerType = TypeVar("HandlerType", bound="EventHandler[Any]")

# handlers
TxnRawHandler = Union[Type[AbstractTransactionHandler], Callable[..., Awaitable[Any]]]
BillRawHandler = Union[Type[AbstractBillHandler], Callable[..., Awaitable[Any]]]
ErrorRawHandler = Union[Type[ErrorHandler], Callable[..., Awaitable[Any]]]

logger = logging.getLogger("glQiwiApi.dispatcher")


class SkipHandler(Exception):
    pass


class CancelHandler(Exception):
    pass


class EventHandler(Generic[Event]):
    """
    Event handler, which executing and working with handlers

    """

    def __init__(
            self,
            handler: Union[Callable[..., Awaitable[Any]], Type[Handler[Event]]],
            *filters: BaseFilter[Event],
    ) -> None:
        self._handler = handler
        self._filters = list(
            filter(lambda f: operator.not_(operator.eq(f, None)), filters)
        )

    async def check_then_execute(self, event: Event, *args: Any) -> Optional[Any]:
        """Check event, apply all filters and then pass on to handler"""
        for f in self._filters:
            if not await f.check(event):
                break
        else:
            return await self._handler(event, *args)
        return None  # hint for mypy


TxnWrappedHandler = Union[EventHandler["Transaction"], EventHandler["WebHook"]]


class HandlerCollection(Generic[Event]):
    def __init__(self, once: bool = True) -> None:
        self.handlers: List[EventHandler[Event]] = []
        self.once = once

    async def notify(self, event: Event, *args: Any) -> None:
        for handler in self.handlers:
            try:
                await handler.check_then_execute(event, *args)
                if self.once:
                    break
            except SkipHandler:
                continue
            except CancelHandler:
                break

    def subscribe(self, handler: EventHandler[Event]) -> None:
        self.handlers.append(handler)

    def unsubscribe(self, handler: EventHandler[Event]) -> None:
        self.handlers.remove(handler)

    def __iter__(self) -> Iterable[EventHandler[Event]]:
        return iter(self.handlers)

    @property
    def is_empty(self) -> bool:
        return len(self.handlers) == 0


class Dispatcher:
    """
    Class, which managing all handlers

    """

    def __init__(self) -> None:
        self.transaction_handlers: HandlerCollection[
            Union["Transaction", "WebHook"]
        ] = HandlerCollection()
        self.bill_handlers: HandlerCollection["Notification"] = HandlerCollection()
        self.error_handlers: HandlerCollection[Exception] = HandlerCollection()

    @no_type_check
    def _wrap_callback_for_error_handling(self, callback):
        @functools.wraps(callback)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return await callback(*args, **kwargs)
            except (SkipHandler, CancelHandler) as ex:
                # reraise exception to handle it on HandlerCollection object
                # and skip or break processing current event
                raise ex
            except Exception as e:
                if not self.error_handlers.is_empty:
                    return await self.error_handlers.notify(e, *args)
                raise e

        return wrapper

    def register_transaction_handler(
            self, event_handler: TxnRawHandler, *filters: TxnFilters
    ) -> None:
        self.transaction_handlers.subscribe(
            cast(  # type: ignore
                TxnWrappedHandler,
                self.wrap_handler(
                    self._wrap_callback_for_error_handling(event_handler),
                    filters,
                    default_filter=TransactionFilter(),
                ),
            )
        )

    def register_bill_handler(
            self, event_handler: BillRawHandler, *filters: BillFilters
    ) -> None:
        self.bill_handlers.subscribe(
            self.wrap_handler(
                self._wrap_callback_for_error_handling(event_handler),
                filters,
                default_filter=BillFilter(),
            )
        )

    def register_error_handler(
            self,
            event_handler: ErrorRawHandler,
            exception: Optional[Union[Type[Exception], Exception]] = None,
            *filters: BaseFilter[Exception],
    ) -> None:
        self.error_handlers.subscribe(
            self.wrap_handler(
                event_handler, filters=filters, default_filter=ErrorFilter(exception)
            )
        )

    @property
    def __all_handlers__(self):  # type: ignore
        """Return all registered handlers except error handlers"""
        return self.bill_handlers, self.transaction_handlers

    @staticmethod
    def wrap_handler(
            event_handler: Union[Callable[..., Awaitable[Any]], Type[Handler[Event]]],
            filters: Optional[
                Tuple[Union[Callable[[Event], bool], BaseFilter[Event]], ...]
            ] = None,
            default_filter: Optional[BaseFilter[Event]] = None,
    ) -> EventHandler[Event]:
        """
        Add new event handler.
        (!) Allows chain addition.

        :param event_handler: event handler, low order function
         which works with events
        :param filters: filter for low order function execution
        :param default_filter: default filter for the handler to determine type
         of update and accurately split handlers by the type of update

        :return: instance of EventHandler
        """
        generated_filters: List[BaseFilter[Event]] = []
        if filters:
            for filter_ in filters:
                if isinstance(filter_, types.FunctionType):
                    generated_filters.append(LambdaBasedFilter(filter_))
                else:
                    generated_filters.append(cast(BaseFilter[Event], filter_))
            if default_filter is not None:
                generated_filters.insert(0, default_filter)
        else:
            generated_filters = [default_filter]  # type: ignore

        return EventHandler(event_handler, *generated_filters)

    def transaction_handler(
            self, *filters: TxnFilters
    ) -> Callable[[TxnRawHandler], TxnRawHandler]:
        def decorator(callback: TxnRawHandler) -> TxnRawHandler:
            self.register_transaction_handler(callback, *filters)
            return callback

        return decorator

    def bill_handler(
            self, *filters: BillFilters
    ) -> Callable[[BillRawHandler], BillRawHandler]:
        def decorator(callback: BillRawHandler) -> BillRawHandler:
            self.register_bill_handler(callback, *filters)
            return callback

        return decorator

    def error_handler(
            self,
            exception: Optional[Union[Type[Exception], Exception]] = None,
            *filters: BaseFilter[Exception],
    ) -> Callable[[ErrorRawHandler], ErrorRawHandler]:
        def decorator(callback: ErrorRawHandler) -> ErrorRawHandler:
            self.register_error_handler(callback, exception, *filters)
            return callback

        return decorator

    async def process_event(self, event: Event) -> None:
        """
        Feed handlers with event.

        :param event: any object that will be translated to handlers
        """
        coroutines = [
            handler_collection.notify(event=event)
            for handler_collection in self.__all_handlers__
        ]
        await asyncio.gather(*coroutines)



from __future__ import annotations

import asyncio
from concurrent.futures import Future
from typing import (
    Any,
    Optional,
    Coroutine,
    Callable,
    Awaitable,
)

from glQiwiApi import types

def run_forever_safe(
    loop: asyncio.AbstractEventLoop,
    callback: Optional[Callable[..., Awaitable[types.N]]] = None,
) -> None: ...
def safe_cancel(
    loop: asyncio.AbstractEventLoop,
    callback: Optional[Callable[..., Awaitable[types.N]]],
) -> None: ...
def _cancel_future(
    loop: asyncio.AbstractEventLoop,
    future: asyncio.Future[types.N],
    executor: types.AnyExecutor,
) -> None: ...
def _stop_loop(loop: asyncio.AbstractEventLoop) -> None: ...
def take_event_loop(set_debug: bool = False) -> asyncio.AbstractEventLoop: ...
def await_sync(future: Future[types.N]) -> types.N: ...
def execute_async_as_sync(
    func: Callable[..., Coroutine[Any, Any, types.N]], *args: object, **kwargs: object
) -> types.N: ...

class async_as_sync:  # NOQA
    _async_shutdown_callback: Optional[Callable[..., Awaitable[types.N]]]
    _sync_shutdown_callback: Optional[Callable[[Any], Any]]
    def __init__(
        self,
        async_shutdown_callback: Optional[Callable[..., Awaitable[types.N]]] = None,
        sync_shutdown_callback: Optional[Callable[[Any], Any]] = None,
    ) -> None: ...
    def __call__(
        self, func: Callable[..., Coroutine[Any, Any, types.N]]
    ) -> Callable[..., types.N]: ...
    def execute_sync_callback(self, result: Any) -> Any: ...



from __future__ import annotations

import logging
from typing import Union, Any, Dict, Collection, Callable

from glQiwiApi import types, QiwiWrapper


class measure_time(object):  # NOQA
    _logger: Union[logging.Logger, None] = None

    def __init__(self, logger: Union[logging.Logger, None] = None) -> None: ...

    def __call__(self, func: types.N) -> types.N: ...

    def _log(self, msg: str, *args: Any) -> None: ...


class allow_response_code:  # NOQA
    status_code: int

    def __init__(self, status_code: int) -> None: ...

    def __call__(self, func: types.N) -> types.N: ...


class override_error_messages:  # NOQA
    status_codes: Dict[int, Dict[str, str]]

    def __init__(self, status_codes: Dict[int, Dict[str, str]]): ...

    def __call__(self, func: types.N) -> types.N: ...


class require:  # NOQA
    _required_attrs: Collection[str]

    def __init__(self, *params: str): ...

    def __call__(self, func: types.N) -> types.N: ...

    def check_is_object_contains_required_attrs(
            self,
            c: QiwiWrapper,
            func: Callable[..., Any]
    ) -> None:
        ...



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel, Field


class QiwiErrorAnswer(BaseModel):
    service_name: str = Field(..., alias="serviceName")
    error_code: str = Field(..., alias="errorCode")
    description: str = Field(..., alias="description")
    user_msg: str = Field(..., alias="userMessage")
    dt: datetime = Field(..., alias="dateTime")
    trace_id: str = Field(..., alias="traceId")

    def construct_error_message(self) -> str:
        return f"Error code: {self.error_code}, description: {self.description}"



from __future__ import annotations

import abc
from typing import Any, Generic, TypeVar, TYPE_CHECKING, Dict, Union

if TYPE_CHECKING:
    from glQiwiApi.qiwi.client import QiwiWrapper  # NOQA  # pragma: no cover
    from glQiwiApi.types.base import Base  # NOQA # pragma: no cover

T = TypeVar("T", bound=Union[Exception, "Base"])


class BaseHandlerMixin(Generic[T]):
    if TYPE_CHECKING:  # pragma: no cover
        event: T


class ClientMixin(Generic[T]):
    if TYPE_CHECKING:  # pragma: no cover
        event: T

    @property
    def client(self) -> "QiwiWrapper":
        return self.event.client

    @property
    def client_data(self) -> Dict[Any, Any]:
        return self.client.config_data


class Handler(abc.ABC, BaseHandlerMixin[T]):
    """Base class for all class-based handlers"""

    def __init__(self, event: T) -> None:
        self.event: T = event

    @abc.abstractmethod
    async def process_event(self) -> Any:  # pragma: no cover  # type: ignore
        raise NotImplementedError

    def __await__(self) -> Any:
        return self.process_event().__await__()



import concurrent.futures as futures
from typing import Union, TypeVar, Optional, Callable, Any

from .amount import Type, CurrencyAmount, PlainAmount
from .particular import WrappedRequestPayload
from .qiwi import (
    Bill,
    BillError,
    Statistic,
    Balance,
    Identification,
    Limit,
    Account,
    QiwiAccountInfo,
    Transaction,
    PaymentInfo,
    OrderDetails,
    RefundBill,
    Polygon,
    Terminal,
    Partner,
    WebHookConfig,
    WebHook,
    Notification,
    P2PKeys,
    CrossRate,
    FreePaymentDetailsFields,
    PaymentMethod,
    Card,
    Restriction,
    Commission,
    InvoiceStatus,
    TransactionType,
    QiwiPayment,
    Source,
    TransactionStatus,
)
from .yoomoney import (
    OperationType,
    OperationDetails,
    PreProcessPaymentResponse,
    Operation,
    Payment,
    IncomingTransaction,
    AccountInfo,
)

ALL_TYPES = Union[
    QiwiAccountInfo,
    Transaction,
    Bill,
    BillError,
    P2PKeys,
    Statistic,
    Limit,
    Account,
    Identification,
    Balance,
    AccountInfo,
    Operation,
    OperationDetails,
    PreProcessPaymentResponse,
    Payment,
    IncomingTransaction,
    Commission,
    PaymentInfo,
    OrderDetails,
    RefundBill,
    Polygon,
    Terminal,
    Partner,
    WebHookConfig,
    WebHook,
    Notification,
    CrossRate,
    FreePaymentDetailsFields,
    PaymentMethod,
    Card,
    Restriction,
    InvoiceStatus,
    QiwiPayment,
]

AnyExecutor = TypeVar(
    "AnyExecutor",
    futures.ThreadPoolExecutor,
    futures.ProcessPoolExecutor,
    Optional[None],
)

FuncT = TypeVar("FuncT", bound=Callable[..., Any])

N = TypeVar("N")

__all__ = (
    "QiwiAccountInfo",
    "Transaction",
    "Bill",
    "BillError",
    "P2PKeys",
    "InvoiceStatus",
    "Statistic",
    "Limit",
    "Account",
    "Identification",
    "Balance",
    "AccountInfo",
    "OperationType",
    "Operation",
    "OperationDetails",
    "PreProcessPaymentResponse",
    "Payment",
    "IncomingTransaction",
    "WrappedRequestPayload",
    "CurrencyAmount",
    "Type",
    "PlainAmount",
    "Commission",
    "ALL_TYPES",
    "PaymentInfo",
    "OrderDetails",
    "RefundBill",
    "Polygon",
    "Terminal",
    "Partner",
    "WebHookConfig",
    "WebHook",
    "Notification",
    "AnyExecutor",
    "FuncT",
    "N",
    "CrossRate",
    "FreePaymentDetailsFields",
    "PaymentMethod",
    "Card",
    "Restriction",
    "TransactionType",
    "QiwiPayment",
    "Source",
    "TransactionStatus"
)



from __future__ import annotations

import ipaddress
import pathlib
import ssl
from datetime import timedelta, datetime
from io import BytesIO
from typing import cast, TYPE_CHECKING, Optional, Iterable, Union, Any, List

from glQiwiApi.ext.exceptions import StateError

if TYPE_CHECKING:
    try:
        from aiogram.types import InputFile  # NOQA
        from cryptography import x509  # NOQA
        from cryptography.hazmat.primitives.asymmetric.rsa import (
            RSAPrivateKeyWithSerialization,  # NOQA
        )  # NOQA
    except ImportError:  # pragma: no cover # type: ignore
        pass


class SSLConfigurator:
    """
    Help you to generate self-signed certificate and load it as ssl.SSLContext
    Firstly, install `cryptography` package => pip install cryptography or poetry add cryptography

    """

    def __init__(
        self,
        hostname: str,  # your host machine ip address
        cert_path: Union[str, pathlib.Path] = "cert.pem",
        pkey_path: Union[str, pathlib.Path] = "pkey.pem",
        ip_addresses: Optional[Iterable[Any]] = None,
        rsa_private_key: Optional["RSAPrivateKeyWithSerialization"] = None,
        public_exponent: int = 65537,
        key_size: int = 2048,
        backend: Optional[Any] = None,
        serial_number: int = 1000,
        expire_days: int = 3650,
    ):
        self.expire_days = expire_days
        self.serial_number = serial_number
        self.backend = backend
        self.key_size = key_size
        self.public_exponent = public_exponent
        self._rsa_private_key = rsa_private_key
        self.ip_addresses = ip_addresses
        self.pkey_path = pkey_path
        self.cert_path = cert_path
        self.hostname = hostname

    def as_ssl_context(self) -> "ssl.SSLContext":
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        try:
            ssl_context.load_cert_chain(str(self.cert_path), str(self.pkey_path))
        except FileNotFoundError:
            raise StateError(
                "There are not certificates in folder."
                " Firstly, you need to call "
                "`SSLConfigurator.generate_self_signed` method."
            )
        return ssl_context

    def _get_self_signed_cert(self) -> bool:
        if isinstance(self.cert_path, pathlib.Path):
            is_cert_is_file = self.cert_path.is_file()
        else:
            is_cert_is_file = pathlib.Path(self.cert_path).is_file()

        if isinstance(self.pkey_path, pathlib.Path):
            is_pkey_is_file = self.pkey_path.is_file()
        else:
            is_pkey_is_file = pathlib.Path(self.pkey_path).is_file()

        return is_cert_is_file and is_pkey_is_file

    def get_or_generate_self_signed(
        self, *name_attributes: "x509.NameAttribute"
    ) -> SSLConfigurator:
        """
        Generates self signed certificate for a hostname, and optional IP addresses.
        Copied from https://gist.github.com/bloodearnest/9017111a313777b9cce5
        """
        if self._get_self_signed_cert():
            return self

        try:
            from cryptography import x509
            from cryptography.x509.oid import NameOID
            from cryptography.hazmat.primitives import hashes
            from cryptography.hazmat.backends import default_backend
            from cryptography.hazmat.primitives import serialization
            from cryptography.hazmat.primitives.asymmetric import rsa
        except ImportError:
            raise ImportError(
                "You need to pip install cryptography for generating self-signed cert"
            )

        # Generate our key
        if self._rsa_private_key is None:
            self._rsa_private_key = rsa.generate_private_key(
                public_exponent=self.public_exponent,
                key_size=self.key_size,
                backend=self.backend or default_backend(),
            )

        name = x509.Name(
            [x509.NameAttribute(NameOID.COMMON_NAME, self.hostname), *name_attributes]
        )

        # best practice seem to be to include the hostname in the SAN,
        # which *SHOULD* mean COMMON_NAME is ignored.
        alt_names: List[x509.GeneralName] = [x509.DNSName(self.hostname)]

        # allow addressing by IP, for when you don't have real DNS (common in most testing scenarios
        if self.ip_addresses:
            for address in self.ip_addresses:
                # openssl wants DNSnames for ips...
                alt_names.append(x509.DNSName(address))
                # ... whereas golang's crypto/tls is stricter, and needs IPAddresses
                # note: older versions of cryptography do not understand ip_address objects
                alt_names.append(x509.IPAddress(ipaddress.ip_address(address)))

        san = x509.SubjectAlternativeName(alt_names)

        # path_len=0 means this cert can only sign itself, not other certs.
        basic_constraints = x509.BasicConstraints(ca=True, path_length=0)
        now = datetime.utcnow()
        cert = (
            x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(self._rsa_private_key.public_key())
            .serial_number(self.serial_number)
            .not_valid_before(now)
            .not_valid_after(now + timedelta(days=self.expire_days))
            .add_extension(basic_constraints, False)
            .add_extension(san, False)
            .sign(self._rsa_private_key, hashes.SHA256(), default_backend())
        )
        cert_pem = cert.public_bytes(
            encoding=cast(serialization.Encoding, serialization.Encoding.PEM)
        )
        key_pem = self._rsa_private_key.private_bytes(
            encoding=cast(serialization.Encoding, serialization.Encoding.PEM),
            format=cast(
                serialization.PrivateFormat,
                serialization.PrivateFormat.TraditionalOpenSSL,
            ),
            encryption_algorithm=serialization.NoEncryption(),
        )

        with open(self.cert_path, "wb") as f1, open(self.pkey_path, "wb") as f2:
            f1.write(cert_pem)
            f2.write(key_pem)
        return self

    def as_input_file(self) -> "InputFile":
        from aiogram.types import InputFile

        with open(self.cert_path, "rb") as f:
            return InputFile(BytesIO(f.read()))



from __future__ import annotations

import abc
from typing import Optional

from glQiwiApi.core.dispatcher.class_based.base import Handler, ClientMixin
from glQiwiApi.types import Bill, PlainAmount


class AbstractBillHandler(Handler[Bill], ClientMixin[Bill], abc.ABC):
    @property
    def bill_id(self) -> str:
        return self.event.bill_id

    @property
    def bill_sum(self) -> PlainAmount:
        return self.event.amount

    @property
    def pay_url(self) -> Optional[str]:
        return self.event.pay_url



from dataclasses import dataclass
from typing import Optional, Any, Dict


@dataclass
class WrappedRequestPayload:
    headers: Dict[Any, Any]
    json: Dict[Any, Any]
    data: Optional[Dict[Any, Any]] = None
    cookies: Optional[Dict[Any, Any]] = None


__all__ = ("WrappedRequestPayload",)



from __future__ import annotations

import re
from collections import namedtuple
from typing import TypeVar, Type, Optional, Any, cast, Pattern

from glQiwiApi.core.dispatcher.webhooks.config import DEFAULT_QIWI_WEBHOOK_PATH

_URL = TypeVar("_URL", bound="WebhookURL")

# Url without port e.g. https://127.0.0.1/ or https://website.com/
HOST_REGEX: Pattern[str] = re.compile(
    r"^(http(s?)://)?"
    r"(((www\.)?[a-zA-Z0-9.\-_]+"
    r"(\.[a-zA-Z]{2,6})+)|"
    r"(\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)"
    r"{3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b))"
    r"(/[a-zA-Z0-9_\-\s./?%#&=]*)?$"
)


class WebhookURL(
    namedtuple(
        "WebhookURL",
        ["host", "webhook_path", "port"],
    )
):
    host: str
    webhook_path: Optional[str]
    port: Optional[int]

    @classmethod
    def create(
            cls: Type[_URL],
            host: str,
            port: Optional[int] = None,
            webhook_path: Optional[str] = None,
    ) -> _URL:
        return cls(
            host=cls._assert_host(host, param_name="host"),
            webhook_path=cls._assert_str(
                webhook_path,
                param_name="webhook_path"
            ),
            port=cls._assert_int(port, param_name="port"),
        )

    @classmethod
    def _assert_int(cls, v: Optional[Any], *, param_name: str) -> Optional[int]:
        if v is None:
            return v

        if not isinstance(v, int):
            raise TypeError("%s must be integer" % param_name)
        return v

    @classmethod
    def _assert_str(
            cls,
            v: Optional[Any],
            *,
            param_name: str
    ) -> Optional[str]:
        if v is None:
            return v

        if not isinstance(v, str):
            raise TypeError("%s must be a string" % param_name)
        return v

    @classmethod
    def _assert_host(cls, v: Any, *, param_name: str) -> str:
        if not re.match(HOST_REGEX, v):
            raise TypeError(
                "%s must be like https://127.0.0.1/ or https://website.com/"
                % param_name
            )
        return cast(str, v)

    def _doesnt_contains_slash(self) -> bool:
        return not (self.host.endswith("/") and self.webhook_path.startswith("/"))  # type: ignore

    def render(self) -> str:
        host = self.host
        if self.webhook_path is None:
            # Here we use `DEFAULT_QIWI_WEBHOOK_PATH` instead of DEFAULT_QIWI_BILLS_WEBHOOK_PATH
            # because the second you need to register directly in QIWI P2P API and it's no need to build url to it
            self.webhook_path = DEFAULT_QIWI_WEBHOOK_PATH
        if self.port is not None:
            host += f":{self.port}"
        if self._doesnt_contains_slash():
            host += "/"
        return f"{host}{self.webhook_path}"



from typing import List

from pydantic import Field

from glQiwiApi.types.base import ExtraBase
from glQiwiApi.types.amount import CurrencyAmount


class Statistic(ExtraBase):
    """object: Statistic"""

    incoming: List[CurrencyAmount] = Field(alias="incomingTotal")
    out: List[CurrencyAmount] = Field(alias="outgoingTotal")


__all__ = ["Statistic"]



import abc
from typing import Dict, Any


class SingletonABCMeta(abc.ABCMeta):
    """
    Abstract singleton metaclass, using for routers because in class methods
    it's not possible to get the router object,
    so we need singleton to get the same instances of routers

    """

    _instances: Dict[Any, Any] = {}

    def __call__(cls, *args, **kwargs):  # type: ignore
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class AbstractRouter(metaclass=SingletonABCMeta):
    def __init__(self) -> None:
        self.config = self.setup_config()
        self.routes = self.setup_routes()

    @abc.abstractmethod
    def setup_config(self) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def setup_routes(self) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_url_kwargs(url_: str, **kwargs: Any) -> str:
        try:
            return url_.format(**kwargs)
        except KeyError:
            raise TypeError("Bad kwargs for url assembly")

    @abc.abstractmethod
    def build_url(self, api_method: str, **kwargs: Any) -> str:
        """Implementation of building url"""
        raise NotImplementedError



from __future__ import annotations

import sys
from typing import Any, AsyncContextManager


class ModuleNotInstalledException(Exception):
    pass


class EmptyCls(object):
    def __init__(self) -> None:
        raise ModuleNotInstalledException()


try:
    import aiofiles
except ImportError:
    class aiofiles_compat:
        def open(self, *args: Any, **kwargs: Any) -> AsyncContextManager[Any]:
            raise ModuleNotInstalledException(
                "Module aiofiles not installed and you can't use it's "
                "functionality till you install this module."
            )
    aiofiles = aiofiles_compat()  # type: ignore

try:
    import orjson as json
except (ImportError, ModuleNotFoundError):
    import json  # type: ignore

try:
    from aiogram import Dispatcher
    from aiogram.types import InputFile
except (ModuleNotFoundError, ImportError):
    Dispatcher = EmptyCls
    InputFile = EmptyCls

if sys.version_info >= (3, 8):
    from typing import Literal as Literal
    from typing import Final as Final
else:
    from typing_extensions import Literal, Final



from __future__ import annotations

import time
from copy import deepcopy
from typing import Any, Dict

from glQiwiApi.core.abc.router import AbstractRouter
from glQiwiApi.types import WrappedRequestPayload
from glQiwiApi.utils.mypy_hacks import lru_cache

__all__ = ("get_config", "QiwiRouter", "QiwiKassaRouter", "QiwiApiMethods")


@lru_cache()
def get_config() -> QiwiConfig:
    return QiwiConfig()


class QiwiRouter(AbstractRouter):
    """The class that instructs to build the right paths except for QIWI P2P API"""

    __head__ = "https://edge.qiwi.com"

    def setup_routes(self) -> Any:
        return QiwiApiMethods()

    @lru_cache(typed=True)
    def build_url(self, api_method: str, **kwargs: Any) -> str:
        pre_build_url = self.__head__ + api_method
        return super()._format_url_kwargs(pre_build_url, **kwargs)

    def setup_config(self) -> QiwiConfig:
        return get_config()

    def generate_default_headers(self) -> Dict[Any, Any]:
        return deepcopy(self.config.DEFAULT_QIWI_HEADERS)


class QiwiKassaRouter(QiwiRouter):
    """QIWI P2P router"""

    __head__ = "https://api.qiwi.com/partner/bill/v1/bills"


class QiwiApiMethods:
    # Qiwi Wallet
    TO_CARD: str = "/sinap/api/v2/terms/{privat_card_id}"
    GET_BALANCE: str = "/funding-sources/v2/persons/{phone_number}/accounts"
    TRANSACTIONS: str = "/payment-history/v2/persons/{stripped_number}/payments"
    TO_WALLET: str = "/sinap/api/v2/terms/99/payments"
    TRANSACTION_INFO: str = "/payment-history/v1/transactions/{transaction_id}"
    CHECK_RESTRICTION: str = (
        "/person-profile/v1/persons/{phone_number}/status/restrictions"
    )
    GET_IDENTIFICATION: str = "/identification/v1/persons/{phone_number}/identification"
    GET_LIMITS: str = "/qw-limits/v1/persons/{stripped_number}/actual-limits"
    GET_LIST_OF_CARDS: str = "/cards/v1/cards"
    AUTHENTICATE: str = "/identification/v1/persons/{stripped_number}/identification"
    GET_BILLS: str = "/checkout-api/api/bill/search"
    GET_RECEIPT: str = "/payment-history/v1/transactions/{transaction_id}/cheque/file"
    COMMISSION: str = "/sinap/providers/{special_code}/onlineCommission"
    ACCOUNT_INFO: str = "/person-profile/v1/profile/current"
    FETCH_STATISTICS: str = (
        "/payment-history/v2/persons/{stripped_number}/payments/total"
    )
    LIST_OF_BALANCES: str = "/funding-sources/v2/persons/{stripped_number}/accounts"
    CREATE_NEW_BALANCE: str = "/funding-sources/v2/persons/{stripped_number}/accounts"
    AVAILABLE_BALANCES: str = (
        "/funding-sources/v2/persons/{stripped_number}/accounts/offer"
    )
    SET_DEFAULT_BALANCE: str = (
        "/funding-sources/v2/persons/{stripped_number}/accounts/{currency_alias}"
    )
    BUY_QIWI_MASTER: str = "/sinap/api/v2/terms/28004/payments"
    CONFIRM_QIWI_MASTER: str = (
        "/cards/v2/persons/{stripped_number}/orders/{order_id}/submit"
    )
    BUY_QIWI_CARD = "/sinap/test_core/v2/terms/32064/payments"
    CARDS_QIWI_MASTER: str = "/cards/v1/cards/?vas-alias=qvc-master"
    PRE_QIWI_REQUEST: str = "/cards/v2/persons/{number}/orders"
    GET_CROSS_RATES: str = "/sinap/crossRates"
    SPECIAL_PAYMENT: str = "/sinap/api/v2/terms/1717/payments"

    # Qiwi P2P
    CREATE_P2P_BILL: str = "/{bill_id}"
    CHECK_P2P_BILL_STATUS: str = "/{bill_id}"
    REJECT_P2P_BILL: str = "/{bill_id}/reject"
    REFUND_BILL: str = "/{bill_id}/refunds/{refund_id}"
    CREATE_P2P_KEYS: str = "/widgets-api/api/p2p/protected/keys/create"

    # Webhooks
    REG_WEBHOOK: str = "/payment-notifier/v1/hooks"
    GET_CURRENT_WEBHOOK: str = "/payment-notifier/v1/hooks/active"
    SEND_TEST_NOTIFICATION: str = "/payment-notifier/v1/hooks/test"
    GET_WEBHOOK_SECRET: str = "/payment-notifier/v1/hooks/{hook_id}/key"
    DELETE_CURRENT_WEBHOOK: str = "/payment-notifier/v1/hooks/{hook_id}"
    CHANGE_WEBHOOK_SECRET: str = "/payment-notifier/v1/hooks/{hook_id}/newkey"


class QiwiConfig:
    def __init__(self) -> None:
        self.DEFAULT_QIWI_HEADERS = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer {token}",
            "Host": "edge.test_qiwi.com",
        }
        self.P2P_QIWI_HEADERS = {
            "Authorization": "Bearer {token}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.ERROR_CODE_MESSAGES = {
            400: "Query syntax error (invalid data format). Can be related to wrong arguments,"
                 " that you have passed to method",
            401: "Invalid token or API token was expired",
            403: "No permission for this request(API token has insufficient permissions)",
            404: "Object was not found or there are no objects with the specified characteristics",
            423: "Too many requests, the service is temporarily unavailable",
            422: "The domain / subnet / host is incorrectly specified"
                 "webhook (in the new_url parameter for the webhook URL),"
                 "the hook type or transaction type is incorrectly specified,"
                 "an attempt to create a hook if there is one already created",
            405: "Error related to the type of API request, contact the developer or open an issue",
            500: "Internal service error",
            0: "An error related to using a proxy or library problems",
        }
        self.QIWI_MASTER = {
            "id": str(int(time.time() * 1000)),
            "sum": {"amount": 2999, "currency": "643"},
            "paymentMethod": {"type": "Account", "accountId": "643"},
            "comment": "test",
            "fields": {"account": "", "vas_alias": "qvc-master"},
        }
        self.QIWI_TO_CARD: WrappedRequestPayload = WrappedRequestPayload(
            json={
                "id": str(int(time.time() * 1000)),
                "sum": {"amount": "", "currency": "643"},
                "paymentMethod": {"type": "Account", "accountId": "643"},
                "fields": {"account": ""},
            },
            headers=self.DEFAULT_QIWI_HEADERS,
        )

        self.P2P_DATA: WrappedRequestPayload = WrappedRequestPayload(
            json={
                "amount": {"currency": "RUB", "value": "{amount}"},
                "expirationDateTime": "",
                "comment": "{comment}",
                "customFields": {
                    "paySourcesFilter": "qw",
                    "themeCode": "Yvan-YKaSh",
                },
            },
            headers=self.P2P_QIWI_HEADERS,
        )

        self.QIWI_TO_WALLET: WrappedRequestPayload = WrappedRequestPayload(
            json={
                "id": str(int(time.time() * 1000)),
                "sum": {"amount": "", "currency": ""},
                "paymentMethod": {"type": "Account", "accountId": "643"},
                "comment": "",
                "fields": {"account": ""},
            },
            headers=self.DEFAULT_QIWI_HEADERS,
        )
        self.COMMISSION_DATA: WrappedRequestPayload = WrappedRequestPayload(
            json={
                "account": "",
                "paymentMethod": {"type": "Account", "accountId": "643"},
                "purchaseTotals": {"total": {"amount": "", "currency": "643"}},
            },
            headers=self.DEFAULT_QIWI_HEADERS,
        )
        self.DEFAULT_LIMIT_TYPES = [
            "TURNOVER",
            "REFILL",
            "PAYMENTS_P2P",
            "PAYMENTS_PROVIDER_INTERNATIONALS",
            "PAYMENTS_PROVIDER_PAYOUT",
            "WITHDRAW_CASH",
        ]



from __future__ import annotations

import ipaddress
import typing
from dataclasses import dataclass

DEFAULT_QIWI_WEBHOOK_PATH = "/webhooks/qiwi/"
DEFAULT_QIWI_ROUTER_NAME = "QIWI"

DEFAULT_QIWI_BILLS_WEBHOOK_PATH = "/webhooks/qiwi/bills/"
DEFAULT_QIWI_BILLS_ROUTER_NAME = "QIWI_BILLS"

RESPONSE_TIMEOUT = 55

ALLOWED_IPS = {
    ipaddress.IPv4Network("79.142.16.0/20"),
    ipaddress.IPv4Network("195.189.100.0/22"),
    ipaddress.IPv4Network("91.232.230.0/23"),
    ipaddress.IPv4Network("91.213.51.0/24"),
}


@dataclass(frozen=True)
class Path:
    bill_path: typing.Optional[str] = None
    transaction_path: typing.Optional[str] = None


__all__ = (
    "DEFAULT_QIWI_WEBHOOK_PATH",
    "DEFAULT_QIWI_ROUTER_NAME",
    "DEFAULT_QIWI_BILLS_WEBHOOK_PATH",
    "DEFAULT_QIWI_BILLS_ROUTER_NAME",
    "RESPONSE_TIMEOUT",
    "ALLOWED_IPS",
    "Path",
)



import base64
import hashlib
import hmac
from datetime import datetime
from typing import Optional, Dict, Any, List, cast

from pydantic import Field, root_validator

from glQiwiApi.types.base import Base, HashableBase
from glQiwiApi.types.amount import HashableSum
from glQiwiApi.types.exceptions import WebhookSignatureUnverified


def _get_sign_field(dictionary: Dict[Any, Any], nested_keys_list: List[str]) -> Any:
    """
    Recursively iter for nested_keys_list and get nested element.

    For example, imagine that we have list(nested_keys_list) like ["x", "y", "z"],
    x and y is a dictionaries and z is our desired value. So, it's not trivial task without recursion.

    Also, imagine that our nested dict(dictionary) is {"x": {"y": "some_value"}} or {"x": {"y": {"z": "some_value"}}}.
    So, in first example z is already value, in second z is a dict with the desired value and with this help function we
    can process both of this cases.

    @param dictionary:
    @param nested_keys_list:
    @return:
    """
    current = dictionary.get(nested_keys_list.pop(0))
    if isinstance(current, dict):
        return _get_sign_field(current, nested_keys_list)
    else:
        return current


class WebhookPayment(HashableBase):
    """Scheme of webhook payment object"""

    account: str = Field(..., alias="account")
    comment: str = Field(..., alias="comment")
    date: datetime = Field(..., alias="date")
    error_code: str = Field(..., alias="errorCode")
    person_id: int = Field(..., alias="personId")
    provider: int = Field(..., alias="provider")
    sign_fields: str = Field(..., alias="signFields")
    status: str = Field(..., alias="status")
    txn_id: str = Field(..., alias="txnId")
    type: str = Field(..., alias="type")
    commission: Optional[HashableSum] = Field(default=None, alias="calc_commission")
    sum: HashableSum = Field(..., alias="sum")
    total: HashableSum = Field(..., alias="total")


class WebHook(HashableBase):
    """Object: WebHook"""

    hash: Optional[str] = Field(default=None, alias="hash")
    hook_id: str = Field(..., alias="hookId")
    message_id: Optional[str] = Field(default=None, alias="messageId")
    test: bool = Field(..., alias="test")
    version: str = Field(..., alias="version")
    payment: Optional[WebhookPayment] = Field(default=None, alias="payment")

    signature: str
    """NOT API field, this signature is generating in `webhook_signature_collector`"""

    @property
    def is_testable(self) -> bool:
        return self.test and not self.payment

    @root_validator(pre=True)
    def webhook_signature_collector(cls, values: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Get webhook signature to confirm it with hash by base64 encoded key.
        payment.signFields is string. e.g. "sum.currency,sum.amount,type,account,txnId".
        So, this string disassembled piece by piece recursively, because each of signField part, that separated by comma
        can be nested by "."

        @param values:
        @return:
        """
        payment = cast(Optional[Dict[Any, Any]], values.get("payment"))
        if payment is None:
            return values

        sign_fields_list = cast(str, payment.get("signFields")).split(",")
        webhook_signature = "|".join(
            str(_get_sign_field(payment, sign_field.split(".")))
            for sign_field in sign_fields_list
        )
        values.update(signature=webhook_signature)
        return values

    def verify_signature(self, webhook_base64_key: str) -> None:
        webhook_key = base64.b64decode(bytes(webhook_base64_key, "utf-8"))
        generated_hash = hmac.new(
            webhook_key, self.signature.encode("utf-8"), hashlib.sha256
        ).hexdigest()
        if generated_hash != self.hash:
            raise WebhookSignatureUnverified()


class HookParameters(Base):
    """hookParameters object"""

    url: str = Field(..., alias="url")


class WebHookConfig(Base):
    """WebHookConfig object"""

    hook_id: str = Field(..., alias="hookId")
    hook_type: str = Field(..., alias="hookType")
    txn_type: str = Field(..., alias="txnType")
    hook_parameters: HookParameters = Field(..., alias="hookParameters")


__all__ = ("WebHookConfig", "WebHook", "WebhookPayment")



from typing import Optional

from pydantic import Field, validator

from glQiwiApi.types.base import Base
from glQiwiApi.types.amount import CurrencyAmount, Type, CurrencyModel
from glQiwiApi.utils.currency_util import Currency


class Account(Base):
    """object: Account"""

    alias: str
    title: str
    fs_alias: str = Field(alias="fsAlias")
    bank_alias: str = Field(alias="bankAlias")
    has_balance: bool = Field(alias="hasBalance")
    balance: Optional[CurrencyAmount] = None
    currency: CurrencyModel
    account_type: Optional[Type] = Field(None, alias="type")
    is_default_account: bool = Field(alias="defaultAccount")

    @validator("currency", pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))


__all__ = "Account"



from __future__ import annotations

import functools
from typing import Any

from glQiwiApi.core.abc.router import AbstractRouter

__all__ = ("YooMoneyRouter", "YooMoneyMethods")


class YooMoneyRouter(AbstractRouter):
    __head__ = "https://yoomoney.ru"

    def setup_routes(self) -> Any:
        return YooMoneyMethods()

    def setup_config(self) -> YooMoneyConfig:
        return YooMoneyConfig()

    @functools.lru_cache()
    def build_url(self, tail: str, **kwargs: Any) -> str:
        pre_build_url = self.__head__ + tail
        return super()._format_url_kwargs(pre_build_url, **kwargs)


class YooMoneyConfig:
    def __init__(self) -> None:
        self.DEFAULT_YOOMONEY_HEADERS = {
            "Host": "yoomoney.ru",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        self.ERROR_CODE_MESSAGES = {
            400: "An error related to the type of request to the api,"
                 "you may have passed an invalid API token",
            401: "A non-existent, expired, or revoked token is specified",
            403: "An operation has been requested for which the token has no rights",
            0: "Proxy error or unexpected server errors",
        }
        self.content_and_auth = {"content_json": True, "auth": True}


class YooMoneyMethods:
    # Operations with token
    BUILD_URL_FOR_AUTH: str = "/oauth/authorize"
    GET_ACCESS_TOKEN: str = "/oauth/token"
    REVOKE_API_TOKEN: str = "/api/revoke"
    ACCOUNT_INFO: str = "/api/account-info"
    # Transactions
    TRANSACTIONS: str = "/api/operation-history"
    TRANSACTION_INFO: str = "/api/operation-details"
    # Payments
    PRE_PROCESS_PAYMENT: str = "/api/request-payment"
    PROCESS_PAYMENT: str = "/api/process-payment"
    ACCEPT_INCOMING_TRANSFER: str = "/api/incoming-transfer-accept"
    INCOMING_TRANSFER_REJECT: str = "/api/incoming-transfer-reject"

    # Forms
    QUICK_PAY_FORM: str = "/quickpay/confirm.xml?"



"""
Managing polling and webhooks
"""
from __future__ import annotations

import abc
import asyncio
import inspect
import logging
from datetime import datetime
from ssl import SSLContext
from typing import (
    Union,
    Optional,
    List,
    Callable,
    TypeVar,
    Any,
    cast,
    Dict,
)

from aiohttp import web

from glQiwiApi.core.constants import (
    DEFAULT_TIMEOUT,
    TIMEOUT_IF_EXCEPTION,
    DEFAULT_APPLICATION_KEY,
)
from glQiwiApi.core.dispatcher.webhooks import server
from glQiwiApi.core.dispatcher.webhooks.config import Path
from glQiwiApi.core.synchronous import adapter
from glQiwiApi.plugins.abc import Pluggable
from glQiwiApi.qiwi.client import QiwiWrapper, MAX_HISTORY_TRANSACTION_LIMIT  # pragma: no cover
from glQiwiApi.types import Transaction
from glQiwiApi.utils.exceptions import NoUpdatesToExecute

__all__ = ["start_webhook", "start_polling"]

T = TypeVar("T")

_CallbackType = Callable[..., Any]

logger = logging.getLogger("glQiwiApi.executor")


def start_webhook(
        client: QiwiWrapper,
        *plugins: Pluggable,
        host: str = "localhost",
        port: int = 8080,
        path: Optional[Path] = None,
        on_startup: Optional[_CallbackType] = None,
        on_shutdown: Optional[_CallbackType] = None,
        app: Optional["web.Application"] = None,
        ssl_context: Optional[SSLContext] = None
) -> None:
    """
    Blocking function that listens for webhooks

    :param client: instance of QiwiWrapper
    :param host: server host
    :param port: server port that open for tcp/ip trans.
    :param path: path for qiwi that will send requests
    :param app: pass web.Application
    :param on_startup: coroutine,which will be executed on startup
    :param on_shutdown: coroutine, which will be executed on shutdown
    :param plugins: List of plugins, that will be executed together with polling.
         For example  builtin TelegramPollingProxy or other
         class, that inherits from BaseProxy, deal with foreign framework/application
         in the background
    :param ssl_context:
    """
    executor = WebhookExecutor(client, *plugins)
    _setup_callbacks(executor, on_startup, on_shutdown)
    executor.start_webhook(
        host=host, port=port, path=path, app=app, ssl_context=ssl_context
    )


def start_polling(
        client: QiwiWrapper,
        *plugins: Pluggable,
        skip_updates: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        on_startup: Optional[_CallbackType] = None,
        on_shutdown: Optional[_CallbackType] = None
) -> None:
    """
    Setup for long-polling mode. Support only `glQiwiApi.types.Transaction` as event.

    :param client: instance of QiwiWrapper
    :param skip_updates:
    :param timeout: timeout of polling in seconds, if the timeout is too small,
         the API can throw an exception
    :param on_startup: function or coroutine,
         which will be executed on startup
    :param on_shutdown: function or coroutine,
         which will be executed on shutdown
    :param plugins: List of plugins, that will be executed together with polling.
         For example  builtin TelegramPollingProxy or other
         class, that inherits from BaseProxy, deal with foreign framework/application
         in the background
    """
    executor = PollingExecutor(
        client, *plugins, timeout=timeout, skip_updates=skip_updates
    )
    _setup_callbacks(executor, on_startup, on_shutdown)
    executor.start_polling()


class BaseExecutor(abc.ABC):
    def __init__(self, client: QiwiWrapper, *plugins: Pluggable,
                 loop: Optional[asyncio.AbstractEventLoop] = None, ) -> None:
        if loop is not None:
            self._loop = loop  # pragma: no cover
        self.client = client
        self._on_startup_calls: List[_CallbackType] = []
        self._on_shutdown_calls: List[_CallbackType] = []
        self._dispatcher = client.dispatcher
        self._plugins = plugins

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        self._loop = cast(
            asyncio.AbstractEventLoop, getattr(self, "_loop", asyncio.get_event_loop())
        )
        # There is a issue, connected with aiohttp.web application, because glQiwiApi wants to have the same interface
        # for polling and webhooks on_startup & on_shutdown callbacks,
        # so we don't use app.on_shutdown.append(...), on the contrary we use self-written
        # system of callbacks, and so aiohttp.web.Application close event loop on shutdown, that's why we need to
        # create new event loop to gracefully execute shutdown callback.
        # It is highly undesirable to delete this line because you will always catch RuntimeWarning and RuntimeError.
        self._create_new_loop_if_closed()
        return self._loop

    def _create_new_loop_if_closed(self) -> None:
        if self._loop.is_closed():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

    async def welcome(self) -> None:
        logger.debug("Executor has started work!")
        for callback in self._on_startup_calls:
            await _inspect_and_execute_callback(callback=callback, client=self.client)  # pragma: no cover

    async def goodbye(self) -> None:
        logger.debug("Goodbye!")
        for callback in self._on_shutdown_calls:
            await _inspect_and_execute_callback(callback=callback, client=self.client)  # pragma: no cover

    def add_shutdown_callback(self, callback: _CallbackType) -> None:
        self._on_shutdown_calls.append(callback)

    def add_startup_callback(self, callback: _CallbackType) -> None:
        self._on_startup_calls.append(callback)

    async def _incline_plugins(self, ctx: Dict[Any, Any]) -> None:
        incline_tasks = [plugin.incline(ctx) for plugin in self._plugins]
        await asyncio.gather(*incline_tasks)

    def _on_shutdown(self) -> None:
        """
        On shutdown, executor gracefully cancel all tasks, close event loop
        and call `close` method to clear resources
        """
        callbacks = [self.goodbye(), self.client.close(), self._shutdown_plugins()]
        self._loop.run_until_complete(asyncio.shield(asyncio.gather(*callbacks)))

    async def _shutdown_plugins(self) -> None:
        logger.debug("Shutting down plugins")
        shutdown_tasks = [asyncio.create_task(plugin.shutdown()) for plugin in self._plugins]
        await asyncio.gather(*shutdown_tasks)


class PollingExecutor(BaseExecutor):
    def __init__(
            self,
            client: QiwiWrapper,
            *plugins: Pluggable,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            timeout: Union[float, int] = DEFAULT_TIMEOUT,
            skip_updates: bool = False,
    ) -> None:
        """

        :param client: instance of BaseWrapper
        :param tg_app: optional proxy to connect aiogram polling/webhook mode
        """
        super(PollingExecutor, self).__init__(client, *plugins, loop=loop)
        self.offset: Optional[int] = None
        self.get_updates_from = datetime.now()
        self._timeout: Union[int, float] = _parse_timeout(timeout)
        self.skip_updates = skip_updates

    def start_polling(self) -> None:
        loop: asyncio.AbstractEventLoop = self.loop
        try:
            loop.run_until_complete(self.welcome())
            loop.create_task(self._start_polling())
            adapter.run_forever_safe(loop=loop)
        except (SystemExit, KeyboardInterrupt):  # pragma: no cover
            # Allow to graceful shutdown
            pass  # pragma: no cover
        finally:
            self._on_shutdown()

    async def _start_polling(self) -> None:
        default_timeout = self._timeout
        while True:
            try:
                await self._try_fetch_new_updates()
                self._set_timeout(default_timeout)
            except Exception as ex:
                self._set_timeout(TIMEOUT_IF_EXCEPTION)
                logger.error("Handle `!r`. Sleeping %s seconds", ex, self._timeout)
            await asyncio.sleep(self._timeout)

    async def _try_fetch_new_updates(self) -> None:
        try:
            updates = await self._fetch_updates()
        except NoUpdatesToExecute:
            return
        if self.offset is None:
            first_update = updates[0]
            self.offset = first_update.id - 1
        logger.debug("Current transaction offset is %d", self.offset)
        await self.process_updates(updates)

    async def _fetch_updates(self) -> List[Transaction]:
        history = await self._get_consistent_history()
        if len(history) == MAX_HISTORY_TRANSACTION_LIMIT:
            logger.debug("History is out of max history transaction limit")
            first_txn_by_date = history[-1]
            self.get_updates_from = first_txn_by_date.date

        if self.skip_updates:
            self.skip_updates = False
            raise NoUpdatesToExecute()
        elif not history:
            logger.debug("History is empty at the time interval %s to %s", self.get_updates_from)
            raise NoUpdatesToExecute()

        return history

    async def _get_consistent_history(self) -> List[Transaction]:
        """
        QIWI API returns history sorted by dates descending, but for "consistency" we have to reverse it,
        which is what this method does.
        """
        history_from_last_to_first = await self.client.transactions(
            start_date=self.get_updates_from,
            end_date=datetime.now()
        )
        history_from_first_to_last = list(reversed(history_from_last_to_first))
        return history_from_first_to_last

    async def process_updates(self, history: List[Transaction]) -> None:
        tasks: List[asyncio.Task[None]] = [
            asyncio.create_task(self._dispatcher.process_event(event))
            for event in history
            if cast(int, self.offset) < event.id
        ]
        if history:
            sorted_history = sorted(history, key=lambda txn: txn.id)
            self.offset = sorted_history[-1].id
        await asyncio.gather(*tasks)

    def _set_timeout(self, exception_timeout: Union[int, float]) -> None:
        self._timeout = exception_timeout


class WebhookExecutor(BaseExecutor):

    def __init__(self, client: QiwiWrapper, *plugins: Pluggable):
        super().__init__(client, *plugins)
        self._application = web.Application()

    def start_webhook(
            self,
            *,
            host: str = "localhost",
            port: int = 8080,
            path: Optional[Path] = None,
            app: Optional[web.Application] = None,
            ssl_context: Optional[SSLContext] = None,
    ) -> None:
        loop: asyncio.AbstractEventLoop = self.loop
        if app is not None:
            self._application = app
        self._application = server.configure_app(
            dispatcher=self._dispatcher,
            app=self._application,
            path=path,
            secret_key=self.client.secret_p2p
        )
        self._add_application_into_client_data()
        try:
            self._incline_plugins(ctx={"app": self._application})
            loop.run_until_complete(self.welcome())
            web.run_app(self._application, host=host, port=port, ssl_context=ssl_context)
        except (KeyboardInterrupt, SystemExit):
            # Allow to graceful shutdown
            pass
        finally:
            self._on_shutdown()

    def _add_application_into_client_data(self) -> None:
        self.client[DEFAULT_APPLICATION_KEY] = self._application

    async def welcome(self) -> None:
        """
        We have to override `welcome` method to gracefully get webhook base64 encoded key to decode webhook entities.
        It allow you to bind webhook in on_startup callback and don't do it separately
        """
        await super(WebhookExecutor, self).welcome()
        # call to `bind_webhook` without args gives us data about
        # the current webhook config and base64 encoded key to decode webhook entities
        hook_config, base64_key = await self.client.bind_webhook()
        self._application["_base64_key"] = base64_key


def _parse_timeout(timeout: Union[float, int]) -> float:
    if isinstance(timeout, float):
        return timeout
    elif isinstance(timeout, int):
        return float(timeout)
    else:
        raise TypeError(f"Timeout must be float or int. You have passed on {type(timeout)}")


async def _inspect_and_execute_callback(client: "QiwiWrapper", callback: _CallbackType) -> None:
    if inspect.iscoroutinefunction(callback):
        await callback(client)
    else:
        callback(client)  # pragma: no cover


def _setup_callbacks(
        executor: BaseExecutor,
        on_startup: Optional[_CallbackType] = None,
        on_shutdown: Optional[_CallbackType] = None,
) -> None:
    """
    Function, which setup callbacks and set it to dispatcher object

    :param executor:
    :param on_startup:
    :param on_shutdown:
    """
    if on_startup is not None:
        executor.add_startup_callback(on_startup)
    if on_shutdown is not None:
        executor.add_shutdown_callback(on_shutdown)  # pragma: no cover



from typing import Optional, Any, Dict

from pydantic import Field

from glQiwiApi.types import CurrencyAmount
from glQiwiApi.types.base import Base


class Fields(Base):
    """object: Fields"""

    account: str


class State(Base):
    """object: State"""

    code: str


class TransactionInfo(Base):
    """object: TransactionInfo"""

    txn_id: int = Field(..., alias="id")
    state: State


class PaymentInfo(Base):
    """object: PaymentInfo"""

    payment_id: int = Field(..., alias="id")
    terms: str
    fields: Optional[Fields] = None
    payment_sum: CurrencyAmount = Field(..., alias="sum")
    source: str
    transaction: Optional[TransactionInfo] = None
    comment: Optional[str] = None


class PaymentMethod(Base):
    type: str = "Account"
    account_id: int = Field(643, alias="accountId")


class QiwiPayment(Base):
    id: int
    sum: CurrencyAmount
    method: PaymentMethod = Field(..., alias="paymentMethod")
    fields: Dict[Any, Any]
    comment: Optional[str] = None


__all__ = ["PaymentInfo", "QiwiPayment"]



from __future__ import annotations

import abc
import asyncio
import http
import time
from asyncio import Task
from dataclasses import dataclass
from typing import Any, Optional, Dict, Union, Tuple, List

UNCACHED = ("https://api.qiwi.com/partner/bill", "/sinap/api/v2/terms/")

INFINITE = float("inf")

ADD_TIME_PLACEHOLDER = "add_time"
VALUE_PLACEHOLDER = "value"


class CacheExpired(Exception):
    pass


class ContainsNonCacheable(Exception):
    pass


def _embed_cache_time(**kwargs: Any) -> Dict[Any, Any]:
    return {
        key: {
            VALUE_PLACEHOLDER: value,
            ADD_TIME_PLACEHOLDER: time.monotonic()
        }
        for key, value in kwargs.items()
    }


class Payload:

    def __init__(self, headers: Optional[Dict[Any, Any]] = None,
                 json: Optional[Dict[Any, Any]] = None,
                 params: Optional[Dict[Any, Any]] = None,
                 data: Optional[Dict[Any, Any]] = None, **kwargs: Any) -> None:
        self.headers = headers
        self.json = json
        self.params = params
        self.data = data

    @classmethod
    def new(cls, kwargs: Dict[Any, Any], args: Tuple[Any, ...]) -> Payload:
        return cls(
            **{k: kwargs.get(k) for k in args if isinstance(kwargs.get(k), dict)}
        )


@dataclass(frozen=True)
class CachedAPIRequest:
    payload: Payload
    response: Any
    method: Union[str, http.HTTPStatus]


class CacheInvalidationStrategy(abc.ABC):

    @abc.abstractmethod
    def process_update(self, **kwargs: Any) -> None:
        pass

    def process_delete(self) -> None:
        pass

    @abc.abstractmethod
    def process_retrieve(self, **kwargs: Any) -> None:
        pass

    async def check_is_contains_similar(self, cached_storage: CacheStorage, item: Any) -> bool:
        raise TypeError


class UnrealizedCacheInvalidationStrategy(CacheInvalidationStrategy):

    def process_update(self, **kwargs: Any) -> None:
        pass

    def process_retrieve(self, **kwargs: Any) -> None:
        pass


class APIResponsesCacheInvalidationStrategy(CacheInvalidationStrategy):
    _validator_criteria = ("params", "json", "data", "headers")

    def __init__(self, cache_time: Union[float, int] = INFINITE):
        self._uncached = ("https://api.qiwi.com/partner/bill", "/sinap/api/v2/terms/")
        self._cache_time = cache_time

    def process_update(self, **kwargs: Any) -> None:
        for key in kwargs.keys():
            if any(key.startswith(coincidence) for coincidence in self._uncached):
                raise ContainsNonCacheable()

    def process_retrieve(self, **kwargs: Any) -> None:
        self._have_cache_expired(**kwargs)

    def _have_cache_expired(self, **kwargs: Dict[str, Union[Any, float]]) -> None:
        for key, value in kwargs.items():
            added_at: float = value[ADD_TIME_PLACEHOLDER]
            elapsed_time_since_adding = time.monotonic() - added_at
            if elapsed_time_since_adding >= self._cache_time:
                raise CacheExpired()

    async def check_is_contains_similar(self, storage: CacheStorage, item: Any) -> bool:
        values = await storage.retrieve_all()
        if not isinstance(item, Payload):
            return item in values
        for value in values:
            if not isinstance(value, CachedAPIRequest):
                continue
            if (
                    value.method == "GET"
                    and value.payload.headers == item.headers  # noqa: W503
                    and value.payload.params == item.params  # noqa: W503
            ):
                return True
            else:
                if value.method == "GET":
                    return False
                keys = (k for k in self._validator_criteria if k != "headers")
                return any(getattr(item, key) == getattr(value.payload, key) for key in keys)
        return False


class CacheStorage(abc.ABC):

    def __init__(self, invalidate_strategy: Optional[CacheInvalidationStrategy] = None):
        if invalidate_strategy is None:
            invalidate_strategy = UnrealizedCacheInvalidationStrategy()
        self._invalidate_strategy: CacheInvalidationStrategy = invalidate_strategy

    @abc.abstractmethod
    def clear(self) -> None:
        pass

    @abc.abstractmethod
    def update(self, **kwargs: Any) -> None:
        pass

    @abc.abstractmethod
    async def retrieve(self, key: str) -> Any:
        pass

    @abc.abstractmethod
    def delete(self, key: str) -> None:
        pass

    @abc.abstractmethod
    async def retrieve_all(self) -> List[Any]:
        ...

    def __getitem__(self, item: Any) -> Any:
        return self.retrieve(item)

    def __setitem__(self, key: str, value: Any) -> Any:
        self.update(**{key: value})


class InMemoryCacheStorage(CacheStorage):
    __slots__ = ("_data", "_invalidate_strategy")

    def __init__(self, invalidate_strategy: Optional[CacheInvalidationStrategy] = None):
        CacheStorage.__init__(self, invalidate_strategy)
        self._data: Dict[Any, Any] = {}
        self._lock = asyncio.Lock()

    def clear(self) -> None:
        self._invalidate_strategy.process_delete()
        self._data.clear()

    async def retrieve_all(self) -> List[Optional[Any]]:
        tasks: List[Task[Optional[Any]]] = [
            asyncio.create_task(self.retrieve(key))
            for key in self._data.keys()
        ]
        return await asyncio.gather(*tasks)  # type: ignore  # noqa

    def update(self, **kwargs: Any) -> None:
        try:
            self._invalidate_strategy.process_update(**kwargs)
        except ContainsNonCacheable:
            return None
        embedded_data: Dict[Any, Any] = _embed_cache_time(**kwargs)
        self._data.update(embedded_data)

    async def retrieve(self, key: str) -> Optional[Any]:
        async with self._lock:
            obj = self._data.get(key)
        if obj is None:
            return obj
        try:
            self._invalidate_strategy.process_retrieve(obj=obj)
            return obj[VALUE_PLACEHOLDER]
        except CacheExpired:
            self.delete(key)
            return None

    def delete(self, key: str) -> None:
        del self._data[key]

    async def contains_similar(self, item: Any) -> bool:
        return await self._invalidate_strategy.check_is_contains_similar(self, item)



from pydantic import Field

from glQiwiApi.types.base import ExtraBase
from glQiwiApi.types.amount import CurrencyAmount


class Commission(ExtraBase):
    provider_id: int = Field(alias="providerId")
    withdraw_sum: CurrencyAmount = Field(alias="withdrawSum")
    enrollment_sum: CurrencyAmount = Field(alias="enrollmentSum")
    qiwi_commission: CurrencyAmount = Field(alias="qwCommission")
    withdraw_to_enrollment_rate: int = Field(alias="withdrawToEnrollmentRate")



from .types import (
    OperationType,
    Operation,
    OperationDetails,
    PreProcessPaymentResponse,
    Payment,
    AccountInfo,
    IncomingTransaction,
)

__all__ = (
    "AccountInfo",
    "OperationType",
    "Operation",
    "OperationDetails",
    "PreProcessPaymentResponse",
    "Payment",
    "IncomingTransaction",
)



from __future__ import annotations

import abc
from typing import Optional

from glQiwiApi.types import WebHook
from glQiwiApi.types.qiwi.webhooks import WebhookPayment
from .base import Handler, ClientMixin


class AbstractWebHookHandler(Handler[WebHook], ClientMixin[WebHook], abc.ABC):
    @property
    def hook_id(self) -> str:
        return self.event.hook_id

    @property
    def payment(self) -> Optional[WebhookPayment]:
        return self.event.payment



from .request_service import RequestService
from .dispatcher.class_based import (
    AbstractBillHandler,
    Handler,
    AbstractTransactionHandler,
    AbstractWebHookHandler,
    ErrorHandler,
)
from .dispatcher.filters import LambdaBasedFilter, BaseFilter
from .dispatcher.webhooks import (
    QiwiBillWebView,
    QiwiWebHookWebView,
    BaseWebHookView,
    server,
)
from .mixins import ContextInstanceMixin
from .abc.wrapper import Wrapper
from .synchronous import (
    async_as_sync,
    execute_async_as_sync,
)

__all__ = (
    "RequestService",
    "ContextInstanceMixin",
    "BaseFilter",
    "LambdaBasedFilter",
    # class-based handlers
    "Handler",
    "AbstractBillHandler",
    "AbstractTransactionHandler",
    "AbstractWebHookHandler",
    "ErrorHandler",
    # synchronous adapters and utils
    "async_as_sync",
    "execute_async_as_sync",
    # webhooks
    "server",
    "QiwiBillWebView",
    "QiwiWebHookWebView",
    "BaseWebHookView",
)
